import os

from packaging import version
import importlib
import importlib.metadata

diffusers_0220_v = version.parse("0.22.0")
diffusers_0240_v = version.parse("0.24.0")
diffusers_0260_v = version.parse("0.26.0")
diffusers_0270_v = version.parse("0.27.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

import diffusers

if diffusers_version < diffusers_0220_v:

    def get_class_obj_and_candidates(
        library_name, class_name, importable_classes, pipelines, is_pipeline_module
    ):
        """Simple helper method to retrieve class object of module as well as potential parent class objects"""
        if is_pipeline_module:
            pipeline_module = getattr(pipelines, library_name)

            class_obj = getattr(pipeline_module, class_name)
            class_candidates = {c: class_obj for c in importable_classes.keys()}
        else:
            if class_name == "UNet2DConditionModel":
                library_name = "onediffx.deep_cache.models.unet_2d_condition"

            if class_name == "UNetSpatioTemporalConditionModel":
                assert diffusers_version >= diffusers_0240_v, (
                    "SVD not support in diffusers-" + str(diffusers_version)
                )
                library_name = (
                    "onediffx.deep_cache.models.unet_spatio_temporal_condition"
                )

            # else we just import it from the library.
            library = importlib.import_module(library_name)

            class_obj = getattr(library, class_name)
            class_candidates = {
                c: getattr(library, c, None) for c in importable_classes.keys()
            }

        return class_obj, class_candidates


else:

    def get_class_obj_and_candidates(
        library_name,
        class_name,
        importable_classes,
        pipelines,
        is_pipeline_module,
        component_name=None,
        cache_dir=None,
    ):
        """Simple helper method to retrieve class object of module as well as potential parent class objects"""
        component_folder = os.path.join(cache_dir, component_name)

        if is_pipeline_module:
            pipeline_module = getattr(pipelines, library_name)

            class_obj = getattr(pipeline_module, class_name)
            class_candidates = {c: class_obj for c in importable_classes.keys()}
        elif os.path.isfile(os.path.join(component_folder, library_name + ".py")):
            # load custom component
            class_obj = get_class_from_dynamic_module(
                component_folder,
                module_file=library_name + ".py",
                class_name=class_name,
            )
            class_candidates = {c: class_obj for c in importable_classes.keys()}
        else:
            # else we just import it from the library.
            if class_name == "UNet2DConditionModel":
                library_name = "onediffx.deep_cache.models.unet_2d_condition"

            if class_name == "UNetSpatioTemporalConditionModel":
                assert diffusers_version >= diffusers_0240_v, (
                    "SVD not support in diffusers-" + str(diffusers_version)
                )
                library_name = (
                    "onediffx.deep_cache.models.unet_spatio_temporal_condition"
                )

            library = importlib.import_module(library_name)
            class_obj = getattr(library, class_name)
            class_candidates = {
                c: getattr(library, c, None) for c in importable_classes.keys()
            }

        return class_obj, class_candidates


ORIGIN_DIFFUDION_PIPELINE = None
ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES = None
ORIGIN_2D_GET_DOWN_BLOCK = None
ORIGIN_2D_GET_UP_BLOCK = None

ORIGIN_3D_GET_DOWN_BLOCK = None
ORIGIN_3D_GET_UP_BLOCK = None

if diffusers_version >= diffusers_0260_v:
    from diffusers.models.unets import unet_2d_condition as diffusers_unet_2d_condition
    from diffusers.models.unets import unet_spatio_temporal_condition as diffusers_unet_spatio_temporal_condition
else:
    from diffusers.models import unet_2d_condition as diffusers_unet_2d_condition
    if diffusers_version >= diffusers_0240_v:
        from diffusers.models import unet_spatio_temporal_condition as diffusers_unet_spatio_temporal_condition


def enable_deep_cache_pipeline():
    global ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES
    global ORIGIN_2D_GET_DOWN_BLOCK
    global ORIGIN_2D_GET_UP_BLOCK
    global ORIGIN_3D_GET_DOWN_BLOCK
    global ORIGIN_3D_GET_UP_BLOCK
    if ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES is None:
        assert ORIGIN_2D_GET_DOWN_BLOCK is None
        assert ORIGIN_2D_GET_UP_BLOCK is None
        if diffusers_version >= diffusers_0240_v:
            assert ORIGIN_3D_GET_DOWN_BLOCK is None
            assert ORIGIN_3D_GET_UP_BLOCK is None
        
        if diffusers_version < diffusers_0270_v:
            ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES = (
                diffusers.pipelines.pipeline_utils.get_class_obj_and_candidates
            )
            diffusers.pipelines.pipeline_utils.get_class_obj_and_candidates = (
                get_class_obj_and_candidates
            )
        else:
            ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES = (
                diffusers.pipelines.pipeline_loading_utils.get_class_obj_and_candidates
            )
            diffusers.pipelines.pipeline_loading_utils.get_class_obj_and_candidates = (
                get_class_obj_and_candidates
            )

        from .unet_2d_blocks import get_down_block as get_2d_down_block

        ORIGIN_2D_GET_DOWN_BLOCK = diffusers_unet_2d_condition.get_down_block
        diffusers_unet_2d_condition.get_down_block = get_2d_down_block

        from .unet_2d_blocks import get_up_block as get_2d_up_block

        ORIGIN_2D_GET_UP_BLOCK = diffusers_unet_2d_condition.get_up_block
        diffusers_unet_2d_condition.get_up_block = get_2d_up_block

        if diffusers_version >= diffusers_0240_v:
            from .unet_3d_blocks import get_down_block as get_3d_down_block

            ORIGIN_3D_GET_DOWN_BLOCK = (
                diffusers_unet_spatio_temporal_condition.get_down_block
            )
            diffusers_unet_spatio_temporal_condition.get_down_block = (
                get_3d_down_block
            )

            from .unet_3d_blocks import get_up_block as get_3d_up_block

            ORIGIN_3D_GET_UP_BLOCK = (
                diffusers_unet_spatio_temporal_condition.get_up_block
            )
            diffusers_unet_spatio_temporal_condition.get_up_block = get_3d_up_block


def disable_deep_cache_pipeline():
    global ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES
    global ORIGIN_2D_GET_DOWN_BLOCK
    global ORIGIN_2D_GET_UP_BLOCK
    global ORIGIN_3D_GET_DOWN_BLOCK
    global ORIGIN_3D_GET_UP_BLOCK
    if ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES is None:
        assert ORIGIN_2D_GET_DOWN_BLOCK is None
        assert ORIGIN_2D_GET_UP_BLOCK is None
        if diffusers_version >= diffusers_0240_v:
            assert ORIGIN_3D_GET_DOWN_BLOCK is None
            assert ORIGIN_3D_GET_UP_BLOCK is None
        return
    if diffusers_version < diffusers_0270_v:
        diffusers.pipelines.pipeline_utils.get_class_obj_and_candidates = (
            ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES
        )
    else:
        diffusers.pipelines.pipeline_loading_utils.get_class_obj_and_candidates = (
            ORIGIN_DIFFUDION_GET_CLC_OBJ_CANDIDATES
        )
    diffusers_unet_2d_condition.get_down_block = ORIGIN_2D_GET_DOWN_BLOCK
    diffusers_unet_2d_condition.get_up_block = ORIGIN_2D_GET_UP_BLOCK
    if diffusers_version >= diffusers_0240_v:
        diffusers_unet_spatio_temporal_condition.get_down_block = ORIGIN_3D_GET_DOWN_BLOCK
        diffusers_unet_spatio_temporal_condition.get_up_block = ORIGIN_3D_GET_UP_BLOCK


__all__ = [
    "enable_deep_cache_pipeline",
    "disable_deep_cache_pipeline",
]



# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List, Optional, Union

from packaging import version
import importlib.metadata

diffusers_0240_v = version.parse("0.24.0")
diffusers_0251_v = version.parse("0.25.1")
diffusers_0263_v = version.parse("0.26.3")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

import numpy as np
import PIL.Image
import torch
from transformers import CLIPImageProcessor, CLIPVisionModelWithProjection

from diffusers.image_processor import VaeImageProcessor
from diffusers.models import AutoencoderKLTemporalDecoder
from diffusers.schedulers import EulerDiscreteScheduler
from diffusers.utils import BaseOutput, logging
from diffusers.utils.torch_utils import randn_tensor

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name

from diffusers import StableVideoDiffusionPipeline as DiffusersStableVideoDiffusionPipeline
from diffusers.pipelines.stable_video_diffusion import (
    StableVideoDiffusionPipelineOutput,
)


from .models.unet_spatio_temporal_condition import UNetSpatioTemporalConditionModel
from .models.fast_unet_spatio_temporal_condition import (
    FastUNetSpatioTemporalConditionModel,
)


from .models.pipeline_utils import enable_deep_cache_pipeline


enable_deep_cache_pipeline()


def _append_dims(x, target_dims):
    """Appends dimensions to the end of a tensor until it has target_dims dimensions."""
    dims_to_append = target_dims - x.ndim
    if dims_to_append < 0:
        raise ValueError(
            f"input has {x.ndim} dims but target_dims is {target_dims}, which is less"
        )
    return x[(...,) + (None,) * dims_to_append]


def tensor2vid(video: torch.Tensor, processor, output_type="np"):
    # Based on:
    # https://github.com/modelscope/modelscope/blob/1509fdb973e5871f37148a4b5e5964cafd43e64d/modelscope/pipelines/multi_modal/text_to_video_synthesis_pipeline.py#L78

    batch_size, channels, num_frames, height, width = video.shape
    outputs = []
    for batch_idx in range(batch_size):
        batch_vid = video[batch_idx].permute(1, 0, 2, 3)
        batch_output = processor.postprocess(batch_vid, output_type)

        outputs.append(batch_output)

    return outputs


class StableVideoDiffusionPipeline(DiffusersStableVideoDiffusionPipeline):
    model_cpu_offload_seq = "image_encoder->unet->vae"
    _callback_tensor_inputs = ["latents"]

    def __init__(
        self,
        vae: AutoencoderKLTemporalDecoder,
        image_encoder: CLIPVisionModelWithProjection,
        unet: UNetSpatioTemporalConditionModel,
        scheduler: EulerDiscreteScheduler,
        feature_extractor: CLIPImageProcessor,
    ):

        self.register_modules(
            vae=vae,
            image_encoder=image_encoder,
            unet=unet,
            scheduler=scheduler,
            feature_extractor=feature_extractor,
        )
        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)
        self.image_processor = VaeImageProcessor(vae_scale_factor=self.vae_scale_factor)

        self.fast_unet = FastUNetSpatioTemporalConditionModel(self.unet)

    @torch.no_grad()
    def __call__(
        self,
        image: Union[PIL.Image.Image, List[PIL.Image.Image], torch.FloatTensor],
        height: int = 576,
        width: int = 1024,
        num_frames: Optional[int] = None,
        num_inference_steps: int = 25,
        min_guidance_scale: float = 1.0,
        max_guidance_scale: float = 3.0,
        fps: int = 7,
        motion_bucket_id: float = 127,
        noise_aug_strength: int = 0.02,
        decode_chunk_size: Optional[int] = None,
        num_videos_per_prompt: Optional[int] = 1,
        generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
        latents: Optional[torch.FloatTensor] = None,
        output_type: Optional[str] = "pil",
        callback_on_step_end: Optional[Callable[[int, int, Dict], None]] = None,
        callback_on_step_end_tensor_inputs: List[str] = ["latents"],
        cache_interval: Optional[int] = 1,
        cache_branch: Optional[int] = None,
        return_dict: bool = True,
    ):
        # 0. Default height and width to unet
        height = height or self.unet.config.sample_size * self.vae_scale_factor
        width = width or self.unet.config.sample_size * self.vae_scale_factor

        num_frames = (
            num_frames if num_frames is not None else self.unet.config.num_frames
        )
        decode_chunk_size = (
            decode_chunk_size if decode_chunk_size is not None else num_frames
        )

        # 1. Check inputs. Raise error if not correct
        self.check_inputs(image, height, width)

        # 2. Define call parameters
        if isinstance(image, PIL.Image.Image):
            batch_size = 1
        elif isinstance(image, list):
            batch_size = len(image)
        else:
            batch_size = image.shape[0]
        device = self._execution_device
        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        if diffusers_version > diffusers_0240_v:
            self._guidance_scale = max_guidance_scale
        else:   
            do_classifier_free_guidance = max_guidance_scale > 1.0

        # 3. Encode input image
        if diffusers_version > diffusers_0240_v:
            image_embeddings = self._encode_image(image, device, num_videos_per_prompt, self.do_classifier_free_guidance)
        else:
            image_embeddings = self._encode_image(
                image, device, num_videos_per_prompt, do_classifier_free_guidance
            )

        # NOTE: Stable Diffusion Video was conditioned on fps - 1, which
        # is why it is reduced here.
        # See: https://github.com/Stability-AI/generative-models/blob/ed0997173f98eaf8f4edf7ba5fe8f15c6b877fd3/scripts/sampling/simple_video_sample.py#L188
        fps = fps - 1

        # 4. Encode input image using VAE
        if diffusers_version > diffusers_0251_v:
            image = self.image_processor.preprocess(image, height=height, width=width).to(device)
            noise = randn_tensor(image.shape, generator=generator, device=device, dtype=image.dtype)
        else:
            image = self.image_processor.preprocess(image, height=height, width=width)
            noise = randn_tensor(
                image.shape, generator=generator, device=image.device, dtype=image.dtype
            )
        image = image + noise_aug_strength * noise

        needs_upcasting = (
            self.vae.dtype == torch.float16 and self.vae.config.force_upcast
        )
        if needs_upcasting:
            self.vae.to(dtype=torch.float32)

        if diffusers_version > diffusers_0251_v:
            image_latents = self._encode_vae_image(
                image,
                device=device,
                num_videos_per_prompt=num_videos_per_prompt,
                do_classifier_free_guidance=self.do_classifier_free_guidance,
            )
        if diffusers_version > diffusers_0240_v:
            image_latents = self._encode_vae_image(image, device, num_videos_per_prompt, self.do_classifier_free_guidance)
        else:
            image_latents = self._encode_vae_image(
                image, device, num_videos_per_prompt, do_classifier_free_guidance
            )
        image_latents = image_latents.to(image_embeddings.dtype)

        # cast back to fp16 if needed
        if needs_upcasting:
            self.vae.to(dtype=torch.float16)

        # Repeat the image latents for each frame so we can concatenate them with the noise
        # image_latents [batch, channels, height, width] ->[batch, num_frames, channels, height, width]
        image_latents = image_latents.unsqueeze(1).repeat(1, num_frames, 1, 1, 1)

        # 5. Get Added Time IDs
        if diffusers_version > diffusers_0240_v:
            added_time_ids = self._get_add_time_ids(
                fps,
                motion_bucket_id,
                noise_aug_strength,
                image_embeddings.dtype,
                batch_size,
                num_videos_per_prompt,
                self.do_classifier_free_guidance,
            )
        else:
            added_time_ids = self._get_add_time_ids(
                fps,
                motion_bucket_id,
                noise_aug_strength,
                image_embeddings.dtype,
                batch_size,
                num_videos_per_prompt,
                do_classifier_free_guidance,
            )
        added_time_ids = added_time_ids.to(device)

        # 4. Prepare timesteps
        self.scheduler.set_timesteps(num_inference_steps, device=device)
        timesteps = self.scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.unet.config.in_channels
        latents = self.prepare_latents(
            batch_size * num_videos_per_prompt,
            num_frames,
            num_channels_latents,
            height,
            width,
            image_embeddings.dtype,
            device,
            generator,
            latents,
        )

        # 7. Prepare guidance scale
        guidance_scale = torch.linspace(
            min_guidance_scale, max_guidance_scale, num_frames
        ).unsqueeze(0)
        guidance_scale = guidance_scale.to(device, latents.dtype)
        guidance_scale = guidance_scale.repeat(batch_size * num_videos_per_prompt, 1)
        guidance_scale = _append_dims(guidance_scale, latents.ndim)

        self._guidance_scale = guidance_scale

        cache_features = None
        interval_seq = list(range(0, num_inference_steps, cache_interval))
        interval_seq = sorted(interval_seq)

        # 8. Denoising loop
        num_warmup_steps = len(timesteps) - num_inference_steps * self.scheduler.order
        self._num_timesteps = len(timesteps)
        with self.progress_bar(total=num_inference_steps) as progress_bar:
            for i, t in enumerate(timesteps):
                # expand the latents if we are doing classifier free guidance
                if diffusers_version > diffusers_0240_v:
                    latent_model_input = torch.cat([latents] * 2) if self.do_classifier_free_guidance else latents
                else:
                    latent_model_input = (
                        torch.cat([latents] * 2) if do_classifier_free_guidance else latents
                    )
                latent_model_input = self.scheduler.scale_model_input(
                    latent_model_input, t
                )

                # Concatenate image_latents over channels dimention
                latent_model_input = torch.cat(
                    [latent_model_input, image_latents], dim=2
                )

                if i in interval_seq:
                    cache_features = None
                    # predict the noise residual
                    noise_pred, cache_features = self.unet(
                        latent_model_input,
                        t,
                        encoder_hidden_states=image_embeddings,
                        added_time_ids=added_time_ids,
                        cache_features=cache_features,
                        cache_branch=cache_branch,
                        return_dict=False,
                    )
                else:
                    noise_pred, cache_features = self.fast_unet(
                        latent_model_input,
                        t,
                        encoder_hidden_states=image_embeddings,
                        added_time_ids=added_time_ids,
                        cache_features=cache_features,
                        cache_branch=cache_branch,
                        return_dict=False,
                    )

                # perform guidance
                if diffusers_version > diffusers_0240_v:
                    if self.do_classifier_free_guidance:
                        noise_pred_uncond, noise_pred_cond = noise_pred.chunk(2)
                        noise_pred = noise_pred_uncond + self.guidance_scale * (noise_pred_cond - noise_pred_uncond)
                else:
                    if do_classifier_free_guidance:
                        noise_pred_uncond, noise_pred_cond = noise_pred.chunk(2)
                        noise_pred = noise_pred_uncond + self.guidance_scale * (
                            noise_pred_cond - noise_pred_uncond
                        )

                # compute the previous noisy sample x_t -> x_t-1
                latents = self.scheduler.step(noise_pred, t, latents).prev_sample

                if callback_on_step_end is not None:
                    callback_kwargs = {}
                    for k in callback_on_step_end_tensor_inputs:
                        callback_kwargs[k] = locals()[k]
                    callback_outputs = callback_on_step_end(self, i, t, callback_kwargs)

                    latents = callback_outputs.pop("latents", latents)

                if i == len(timesteps) - 1 or (
                    (i + 1) > num_warmup_steps and (i + 1) % self.scheduler.order == 0
                ):
                    progress_bar.update()

        if not output_type == "latent":
            # cast back to fp16 if needed
            if needs_upcasting:
                self.vae.to(dtype=torch.float16)
            frames = self.decode_latents(latents, num_frames, decode_chunk_size)
            frames = tensor2vid(frames, self.image_processor, output_type=output_type)
        else:
            frames = latents

        self.maybe_free_model_hooks()

        if not return_dict:
            return frames

        return StableVideoDiffusionPipelineOutput(frames=frames)



from pathlib import Path
from typing import Optional, Union, Dict, Tuple, List
from collections import OrderedDict, defaultdict
from packaging import version

import torch

from onediff.utils import logger

import diffusers
from diffusers.loaders import LoraLoaderMixin

if version.parse(diffusers.__version__) >= version.parse("0.21.0"):
    from diffusers.models.lora import PatchedLoraProjection
else:
    from diffusers.loaders import PatchedLoraProjection


from .utils import (
    _unfuse_lora,
    _set_adapter,
    _delete_adapter,
    _maybe_map_sgm_blocks_to_diffusers,
    is_peft_available,
)
from .text_encoder import load_lora_into_text_encoder
from .unet import load_lora_into_unet

if is_peft_available():
    import peft
is_onediffx_lora_available = version.parse(diffusers.__version__) >= version.parse("0.19.3")


USE_PEFT_BACKEND = False


def load_and_fuse_lora(
    pipeline: LoraLoaderMixin,
    pretrained_model_name_or_path_or_dict: Union[str, Path, Dict[str, torch.Tensor]],
    adapter_name: Optional[str] = None,
    *,
    lora_scale: float = 1.0,
    offload_device="cuda",
    use_cache=False,
    **kwargs,
) -> None:
    if not is_onediffx_lora_available:
        raise RuntimeError(
            "onediffx.lora only supports diffusers of at least version 0.19.3"
        )

    _init_adapters_info(pipeline)
    pipeline._adapter_names.add(adapter_name)
    pipeline._active_adapter_names[adapter_name] = 1.0

    self = pipeline

    if use_cache:
        state_dict, network_alphas = load_state_dict_cached(
            pretrained_model_name_or_path_or_dict, unet_config=self.unet.config, **kwargs,
        )
    else:
        # for diffusers <= 0.20
        if hasattr(LoraLoaderMixin, "_map_sgm_blocks_to_diffusers"):
            orig_func = getattr(LoraLoaderMixin, "_map_sgm_blocks_to_diffusers")
            LoraLoaderMixin._map_sgm_blocks_to_diffusers = _maybe_map_sgm_blocks_to_diffusers
        state_dict, network_alphas = LoraLoaderMixin.lora_state_dict(
            pretrained_model_name_or_path_or_dict, unet_config=self.unet.config, **kwargs,
        )
        if hasattr(LoraLoaderMixin, "_map_sgm_blocks_to_diffusers"):
            LoraLoaderMixin._map_sgm_blocks_to_diffusers = orig_func

    is_correct_format = all("lora" in key for key in state_dict.keys())
    if not is_correct_format:
        raise ValueError("[OneDiffX load_and_fuse_lora] Invalid LoRA checkpoint.")

    # load lora into unet
    load_lora_into_unet(
        self,
        state_dict,
        network_alphas,
        self.unet,
        adapter_name=adapter_name,
        lora_scale=lora_scale,
        offload_device=offload_device,
        use_cache=use_cache,
    )

    # load lora weights into text encoder
    text_encoder_state_dict = {k: v for k, v in state_dict.items() if "text_encoder." in k}
    if len(text_encoder_state_dict) > 0:
        load_lora_into_text_encoder(
            self,
            text_encoder_state_dict,
            network_alphas=network_alphas,
            text_encoder=self.text_encoder,
            prefix="text_encoder",
            lora_scale=lora_scale,
            adapter_name=adapter_name,
            _pipeline=self,
        )

    text_encoder_2_state_dict = {k: v for k, v in state_dict.items() if "text_encoder_2." in k}
    if len(text_encoder_2_state_dict) > 0 and hasattr(self, "text_encoder_2"):
        load_lora_into_text_encoder(
            self,
            text_encoder_2_state_dict,
            network_alphas=network_alphas,
            text_encoder=self.text_encoder_2,
            prefix="text_encoder_2",
            lora_scale=lora_scale,
            adapter_name=adapter_name,
            _pipeline=self,
        )


def unfuse_lora(pipeline: LoraLoaderMixin):
    def _unfuse_lora_apply(m: torch.nn.Module):
        if isinstance(m, (torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d)):
            _unfuse_lora(m)
        elif is_peft_available() and isinstance(
            m, (peft.tuners.lora.layer.Linear, peft.tuners.lora.layer.Conv2d),
        ):
            _unfuse_lora(m.base_layer)

    pipeline._adapter_names.clear()
    pipeline._active_adapter_names.clear()

    pipeline.unet.apply(_unfuse_lora_apply)
    if hasattr(pipeline, "text_encoder"):
        pipeline.text_encoder.apply(_unfuse_lora_apply)
    if hasattr(pipeline, "text_encoder_2"):
        pipeline.text_encoder_2.apply(_unfuse_lora_apply)


def set_and_fuse_adapters(
    pipeline: LoraLoaderMixin,
    adapter_names: Union[List[str], str],
    adapter_weights: Optional[List[float]] = None,
):
    if isinstance(adapter_names, str):
        adapter_names = [adapter_names]

    if adapter_weights is None:
        adapter_weights = [1.0, ] * len(adapter_names)
    elif isinstance(adapter_weights, float):
        adapter_weights = [adapter_weights, ] * len(adapter_names)

    _init_adapters_info(pipeline)
    pipeline._adapter_names |= set(adapter_names)
    pipeline._active_adapter_names = {k: v for k, v in zip(adapter_names, adapter_weights)}

    def set_adapters_apply(m):
        if isinstance(m, (torch.nn.Linear, torch.nn.Conv2d, PatchedLoraProjection)):
            _set_adapter(m, adapter_names, adapter_weights)
        elif is_peft_available() and isinstance(
            m, (peft.tuners.lora.layer.Linear, peft.tuners.lora.layer.Conv2d),
        ):
            _set_adapter(m.base_layer, adapter_names, adapter_weights)

    pipeline.unet.apply(set_adapters_apply)
    if hasattr(pipeline, "text_encoder"):
        pipeline.text_encoder.apply(set_adapters_apply)
    if hasattr(pipeline, "text_encoder_2"):
        pipeline.text_encoder_2.apply(set_adapters_apply)


def delete_adapters(self, adapter_names: Union[List[str], str] = None):
    if adapter_names is None:
        adapter_names = list(self._adapter_names)
    elif isinstance(adapter_names, str):
        adapter_names = [adapter_names]

    for adapter_name in adapter_names:
        self._adapter_names.remove(adapter_name)
        self._active_adapter_names.pop(adapter_name, None)

    def delete_adapters_apply(m):
        if isinstance(m, (torch.nn.Linear, torch.nn.Conv2d, PatchedLoraProjection)):
            _delete_adapter(m, adapter_names)
        elif is_peft_available() and isinstance(
            m, (peft.tuners.lora.layer.Linear, peft.tuners.lora.layer.Conv2d),
        ):
            _delete_adapter(m.base_layer, adapter_names)

    self.unet.apply(delete_adapters_apply)
    if hasattr(self, "text_encoder"):
        self.text_encoder.apply(delete_adapters_apply)
    if hasattr(self, "text_encoder_2"):
        self.text_encoder_2.apply(delete_adapters_apply)


def get_active_adapters(self) -> List[str]:
    if hasattr(self, "_adapter_names"):
        return list(self._active_adapter_names.keys())
    else:
        return []


def _init_adapters_info(self: torch.nn.Module):
    if not hasattr(self, "_adapter_names"):
        setattr(self, "_adapter_names", set())
        setattr(self, "_active_adapter_names", {})


class LRUCacheDict(OrderedDict):
    def __init__(self, capacity):
        super().__init__()
        self.capacity = capacity

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        if len(self) >= self.capacity:
            oldest_key = next(iter(self))
            del self[oldest_key]
        super().__setitem__(key, value)


def load_state_dict_cached(
    lora: Union[str, Path, Dict[str, torch.Tensor]], **kwargs,
) -> Tuple[Dict, Dict]:
    assert isinstance(lora, (str, Path, dict))
    if isinstance(lora, dict):
        state_dict, network_alphas = LoraLoaderMixin.lora_state_dict(lora, **kwargs)
        return state_dict, network_alphas

    global CachedLoRAs
    weight_name = kwargs.get("weight_name", None)

    lora_name = str(lora) + (f"/{weight_name}" if weight_name else "")
    if lora_name in CachedLoRAs:
        logger.debug(f"[OneDiffX Cached LoRA] get cached lora of name: {str(lora_name)}")
        return CachedLoRAs[lora_name]

    state_dict, network_alphas = LoraLoaderMixin.lora_state_dict(lora, **kwargs,)
    CachedLoRAs[lora_name] = (state_dict, network_alphas)
    logger.debug(f"[OneDiffX Cached LoRA] create cached lora of name: {str(lora_name)}")
    return state_dict, network_alphas


CachedLoRAs = LRUCacheDict(100)



from collections import defaultdict
from packaging import version

import torch
import diffusers

if version.parse(diffusers.__version__) >= version.parse("0.22.0"):
    from diffusers.utils import convert_state_dict_to_diffusers
else:
    from .state_dict_utils import convert_state_dict_to_diffusers

if version.parse(diffusers.__version__) >= version.parse("0.24.0"):
    from diffusers.models.lora import (
        text_encoder_attn_modules,
        text_encoder_mlp_modules,
    )
else:
    from diffusers.loaders import text_encoder_attn_modules, text_encoder_mlp_modules
from diffusers.utils import is_accelerate_available

from diffusers.models.modeling_utils import _LOW_CPU_MEM_USAGE_DEFAULT
from onediff.utils import logger

from .utils import fuse_lora, get_adapter_names

USE_PEFT_BACKEND = False

if is_accelerate_available():
    from accelerate.hooks import AlignDevicesHook, CpuOffload, remove_hook_from_module

# The code is mainly referenced from https://github.com/huggingface/diffusers/blob/b09b90e24c7ef0252a1a587939972c2e02d305a6/src/diffusers/loaders/lora.py#L485
def load_lora_into_text_encoder(
    cls,
    state_dict,
    network_alphas,
    text_encoder,
    prefix=None,
    lora_scale=1.0,
    low_cpu_mem_usage=None,
    adapter_name=None,
    _pipeline=None,
):
    """
    This will load and fuse the LoRA layers specified in `state_dict` into `text_encoder`

    Parameters:
        cls (`LoraLoaderMixin`):
            The pipeline to load lora into its text encoder
        state_dict (`dict`):
            A standard state dict containing the lora layer parameters. The key should be prefixed with an
            additional `text_encoder` to distinguish between unet lora layers.
        network_alphas (`Dict[str, float]`):
            See `LoRALinearLayer` for more details.
        text_encoder (`CLIPTextModel`):
            The text encoder model to load the LoRA layers into.
        prefix (`str`):
            Expected prefix of the `text_encoder` in the `state_dict`.
        lora_scale (`float`):
            How much to scale the output of the lora linear layer before it is added with the output of the regular
            lora layer.
        low_cpu_mem_usage (`bool`, *optional*, defaults to `True` if torch version >= 1.9.0 else `False`):
            Speed up model loading only loading the pretrained weights and not initializing the weights. This also
            tries to not use more than 1x model size in CPU memory (including peak memory) while loading the model.
            Only supported for PyTorch >= 1.9.0. If you are using an older version of PyTorch, setting this
            argument to `True` will raise an error.
        adapter_name (`str`, *optional*):
            Adapter name to be used for referencing the loaded adapter model. If not specified, it will use
            `default_{i}` where i is the total number of adapters being loaded.
    """
    low_cpu_mem_usage = (
        low_cpu_mem_usage if low_cpu_mem_usage is not None else _LOW_CPU_MEM_USAGE_DEFAULT
    )

    if adapter_name is None:
        adapter_name = get_adapter_names(text_encoder)

    if hasattr(text_encoder, "adapter_names"):
        if adapter_name in text_encoder.adapter_names:
            raise ValueError(
                f"[OneDiffX load_lora_into_text_encoder] The adapter name {adapter_name} already exists in text_encoder"
            )
        else:
            text_encoder.adapter_name.add(adapter_name)
            text_encoder.active_adapter_name[adapter_name] = 1.0
    else:
        text_encoder.adapter_name = set([adapter_name])
        text_encoder.active_adapter_name = {adapter_name: 1.0}

    # If the serialization format is new (introduced in https://github.com/huggingface/diffusers/pull/2918),
    # then the `state_dict` keys should have `self.unet_name` and/or `self.text_encoder_name` as
    # their prefixes.
    keys = list(state_dict.keys())
    prefix = cls.text_encoder_name if prefix is None else prefix

    # Safe prefix to check with.
    if any(cls.text_encoder_name in key for key in keys):
        # Load the layers corresponding to text encoder and make necessary adjustments.
        text_encoder_keys = [k for k in keys if k.startswith(prefix) and k.split(".")[0] == prefix]
        text_encoder_lora_state_dict = {
            k.replace(f"{prefix}.", ""): v for k, v in state_dict.items() if k in text_encoder_keys
        }

        if len(text_encoder_lora_state_dict) > 0:
            logger.info(f"Loading {prefix}.")
            rank = {}
            text_encoder_lora_state_dict = convert_state_dict_to_diffusers(
                text_encoder_lora_state_dict
            )

            if USE_PEFT_BACKEND:
                # convert state dict
                text_encoder_lora_state_dict = convert_state_dict_to_peft(
                    text_encoder_lora_state_dict
                )

                for name, _ in text_encoder_attn_modules(text_encoder):
                    rank_key = f"{name}.out_proj.lora_B.weight"
                    rank[rank_key] = text_encoder_lora_state_dict[rank_key].shape[1]

                patch_mlp = any(".mlp." in key for key in text_encoder_lora_state_dict.keys())
                if patch_mlp:
                    for name, _ in text_encoder_mlp_modules(text_encoder):
                        rank_key_fc1 = f"{name}.fc1.lora_B.weight"
                        rank_key_fc2 = f"{name}.fc2.lora_B.weight"

                        rank[rank_key_fc1] = text_encoder_lora_state_dict[rank_key_fc1].shape[1]
                        rank[rank_key_fc2] = text_encoder_lora_state_dict[rank_key_fc2].shape[1]
            else:
                for name, _ in text_encoder_attn_modules(text_encoder):
                    rank_key = f"{name}.out_proj.lora_linear_layer.up.weight"
                    rank.update({rank_key: text_encoder_lora_state_dict[rank_key].shape[1]})

                patch_mlp = any(".mlp." in key for key in text_encoder_lora_state_dict.keys())
                if patch_mlp:
                    for name, _ in text_encoder_mlp_modules(text_encoder):
                        rank_key_fc1 = f"{name}.fc1.lora_linear_layer.up.weight"
                        rank_key_fc2 = f"{name}.fc2.lora_linear_layer.up.weight"
                        rank[rank_key_fc1] = text_encoder_lora_state_dict[rank_key_fc1].shape[1]
                        rank[rank_key_fc2] = text_encoder_lora_state_dict[rank_key_fc2].shape[1]

            # group text encoder lora state_dict
            te_lora_grouped_dict = defaultdict(dict)
            for k, v in text_encoder_lora_state_dict.items():
                name_elems = k.split(".")
                module_name, lora_module_name = (
                    ".".join(name_elems[:6]),
                    ".".join(name_elems[6:]),
                )
                te_lora_grouped_dict[module_name][lora_module_name] = v

            if network_alphas is not None:
                alpha_keys = [
                    k
                    for k in network_alphas.keys()
                    if k.startswith(prefix) and k.split(".")[0] == prefix
                ]
                network_alphas = {
                    k.replace(f"{prefix}.", ""): v
                    for k, v in network_alphas.items()
                    if k in alpha_keys
                }

            if USE_PEFT_BACKEND:
                from peft import LoraConfig

                lora_config_kwargs = get_peft_kwargs(
                    rank, network_alphas, text_encoder_lora_state_dict, is_unet=False
                )

                lora_config = LoraConfig(**lora_config_kwargs)

                # adapter_name
                if adapter_name is None:
                    adapter_name = get_adapter_name(text_encoder)

                (
                    is_model_cpu_offload,
                    is_sequential_cpu_offload,
                ) = cls._optionally_disable_offloading(_pipeline)

                # inject LoRA layers and load the state dict
                # in transformers we automatically check whether the adapter name is already in use or not
                text_encoder.load_adapter(
                    adapter_name=adapter_name,
                    adapter_state_dict=text_encoder_lora_state_dict,
                    peft_config=lora_config,
                )

                # scale LoRA layers with `lora_scale`
                scale_lora_layers(text_encoder, weight=lora_scale)
            else:

                network_alphas = {} if network_alphas is None else network_alphas
                is_network_alphas_populated = len(network_alphas) > 0

                for name, attn_module in text_encoder_attn_modules(text_encoder):
                    query_alpha = network_alphas.pop(name + ".to_q_lora.down.weight.alpha", None)
                    key_alpha = network_alphas.pop(name + ".to_k_lora.down.weight.alpha", None)
                    value_alpha = network_alphas.pop(name + ".to_v_lora.down.weight.alpha", None)
                    out_alpha = network_alphas.pop(name + ".to_out_lora.down.weight.alpha", None)

                    if isinstance(rank, dict):
                        current_rank = rank.pop(f"{name}.out_proj.lora_linear_layer.up.weight")
                    else:
                        current_rank = rank

                    fuse_lora(
                        attn_module.q_proj,
                        te_lora_grouped_dict.pop(f"{name}.q_proj"),
                        lora_scale,
                        query_alpha,
                        current_rank,
                        adapter_name=adapter_name,
                        prefix="lora_linear_layer",
                    )
                    fuse_lora(
                        attn_module.k_proj,
                        te_lora_grouped_dict.pop(f"{name}.k_proj"),
                        lora_scale,
                        key_alpha,
                        current_rank,
                        adapter_name=adapter_name,
                        prefix="lora_linear_layer",
                    )
                    fuse_lora(
                        attn_module.v_proj,
                        te_lora_grouped_dict.pop(f"{name}.v_proj"),
                        lora_scale,
                        value_alpha,
                        current_rank,
                        adapter_name=adapter_name,
                        prefix="lora_linear_layer",
                    )
                    fuse_lora(
                        attn_module.out_proj,
                        te_lora_grouped_dict.pop(f"{name}.out_proj"),
                        lora_scale,
                        out_alpha,
                        current_rank,
                        adapter_name=adapter_name,
                        prefix="lora_linear_layer",
                    )

                if patch_mlp:
                    for name, mlp_module in text_encoder_mlp_modules(text_encoder):
                        fc1_alpha = network_alphas.pop(
                            name + ".fc1.lora_linear_layer.down.weight.alpha", None
                        )
                        fc2_alpha = network_alphas.pop(
                            name + ".fc2.lora_linear_layer.down.weight.alpha", None
                        )

                        current_rank_fc1 = rank.pop(f"{name}.fc1.lora_linear_layer.up.weight")
                        current_rank_fc2 = rank.pop(f"{name}.fc2.lora_linear_layer.up.weight")

                        fuse_lora(
                            mlp_module.fc1,
                            te_lora_grouped_dict.pop(f"{name}.fc1"),
                            lora_scale,
                            fc1_alpha,
                            current_rank_fc1,
                            adapter_name=adapter_name,
                            prefix="lora_linear_layer",
                        )
                        fuse_lora(
                            mlp_module.fc2,
                            te_lora_grouped_dict.pop(f"{name}.fc2"),
                            lora_scale,
                            fc2_alpha,
                            current_rank_fc2,
                            adapter_name=adapter_name,
                            prefix="lora_linear_layer",
                        )

                if is_network_alphas_populated and len(network_alphas) > 0:
                    raise ValueError(
                        f"The `network_alphas` has to be empty at this point but has the following keys \n\n {', '.join(network_alphas.keys())}"
                    )

                is_pipeline_offloaded = _pipeline is not None and any(
                    isinstance(c, torch.nn.Module) and hasattr(c, "_hf_hook")
                    for c in _pipeline.components.values()
                )
                if is_pipeline_offloaded and low_cpu_mem_usage:
                    low_cpu_mem_usage = True
                    logger.info(
                        f"Pipeline {_pipeline.__class__} is offloaded. Therefore low cpu mem usage loading is forced."
                    )
                unexpected_keys = list(te_lora_grouped_dict.keys())
                if len(unexpected_keys) != 0:
                    raise ValueError(
                        f"failed to load text encoder state dict, unexpected keys: {unexpected_keys}"
                    )

                # <Unsafe code
                # We can be sure that the following works as all we do is change the dtype and device of the text encoder
                # Now we remove any existing hooks to
                is_model_cpu_offload = False
                is_sequential_cpu_offload = False
                if _pipeline is not None:
                    for _, component in _pipeline.components.items():
                        if isinstance(component, torch.nn.Module):
                            if hasattr(component, "_hf_hook"):
                                is_model_cpu_offload = isinstance(
                                    getattr(component, "_hf_hook"), CpuOffload
                                )
                                is_sequential_cpu_offload = isinstance(
                                    getattr(component, "_hf_hook"), AlignDevicesHook
                                )
                                logger.info(
                                    "Accelerate hooks detected. Since you have called `load_lora_weights()`, the previous hooks will be first removed. Then the LoRA parameters will be loaded and the hooks will be applied again."
                                )
                                remove_hook_from_module(
                                    component, recurse=is_sequential_cpu_offload
                                )

            text_encoder.to(device=text_encoder.device, dtype=text_encoder.dtype)

            # Offload back.
            if is_model_cpu_offload:
                _pipeline.enable_model_cpu_offload()
            elif is_sequential_cpu_offload:
                _pipeline.enable_sequential_cpu_offload()
            # Unsafe code />



from typing import Any, Dict, Optional, Tuple, Union
import types

from packaging import version
import importlib.metadata

diffusers_0260_v = version.parse("0.26.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

import torch
import diffusers
from diffusers.utils import is_torch_version


if diffusers_version >= diffusers_0260_v:
    from diffusers.models.unets import unet_3d_blocks as diffusers_unet_3d_blocks
else:
    from diffusers.models import unet_3d_blocks as diffusers_unet_3d_blocks


class DownBlockSpatioTemporal(diffusers_unet_3d_blocks.DownBlockSpatioTemporal):
    def forward(
        self,
        hidden_states: torch.FloatTensor,
        temb: Optional[torch.FloatTensor] = None,
        image_only_indicator: Optional[torch.Tensor] = None,
        exist_module_idx: Optional[int] = None,
    ) -> Tuple[torch.FloatTensor, Tuple[torch.FloatTensor, ...]]:
        # print("exist_module_idx:", exist_module_idx)
        output_states = ()
        for resnet in self.resnets:
            if self.training and self.gradient_checkpointing:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs)

                    return custom_forward

                if is_torch_version(">=", "1.11.0"):
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet),
                        hidden_states,
                        temb,
                        image_only_indicator,
                        use_reentrant=False,
                    )
                else:
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet),
                        hidden_states,
                        temb,
                        image_only_indicator,
                    )
            else:
                hidden_states = resnet(
                    hidden_states, temb, image_only_indicator=image_only_indicator,
                )

            output_states = output_states + (hidden_states,)

            if (
                exist_module_idx is not None
                and exist_module_idx == len(output_states) - 1
            ):
                return hidden_states, output_states

        if self.downsamplers is not None:
            for downsampler in self.downsamplers:
                hidden_states = downsampler(hidden_states)

            output_states = output_states + (hidden_states,)

        return hidden_states, output_states


class CrossAttnDownBlockSpatioTemporal(diffusers_unet_3d_blocks.CrossAttnDownBlockSpatioTemporal):
    def forward(
        self,
        hidden_states: torch.FloatTensor,
        temb: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        image_only_indicator: Optional[torch.Tensor] = None,
        exist_module_idx: Optional[int] = None,
    ) -> Tuple[torch.FloatTensor, Tuple[torch.FloatTensor, ...]]:
        # print("exist_module_idx:", exist_module_idx)
        output_states = ()

        blocks = list(zip(self.resnets, self.attentions))
        for resnet, attn in blocks:
            if self.training and self.gradient_checkpointing:  # TODO

                def create_custom_forward(module, return_dict=None):
                    def custom_forward(*inputs):
                        if return_dict is not None:
                            return module(*inputs, return_dict=return_dict)
                        else:
                            return module(*inputs)

                    return custom_forward

                ckpt_kwargs: Dict[str, Any] = {
                    "use_reentrant": False
                } if is_torch_version(">=", "1.11.0") else {}
                hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(resnet),
                    hidden_states,
                    temb,
                    image_only_indicator,
                    **ckpt_kwargs,
                )

                hidden_states = attn(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    image_only_indicator=image_only_indicator,
                    return_dict=False,
                )[0]
            else:
                hidden_states = resnet(
                    hidden_states, temb, image_only_indicator=image_only_indicator,
                )
                hidden_states = attn(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    image_only_indicator=image_only_indicator,
                    return_dict=False,
                )[0]

            output_states = output_states + (hidden_states,)
            if (
                exist_module_idx is not None
                and exist_module_idx == len(output_states) - 1
            ):
                return hidden_states, output_states

        if self.downsamplers is not None:
            for downsampler in self.downsamplers:
                hidden_states = downsampler(hidden_states)

            output_states = output_states + (hidden_states,)

        return hidden_states, output_states


class UpBlockSpatioTemporal(diffusers_unet_3d_blocks.UpBlockSpatioTemporal):
    def forward(
        self,
        hidden_states: torch.FloatTensor,
        res_hidden_states_tuple: Tuple[torch.FloatTensor, ...],
        temb: Optional[torch.FloatTensor] = None,
        image_only_indicator: Optional[torch.Tensor] = None,
        enter_module_idx: Optional[int] = None,
    ) -> torch.FloatTensor:
        # print("enter_module_idx:", enter_module_idx)
        prv_f = []
        for idx, resnet in enumerate(self.resnets):
            if enter_module_idx is not None and idx < enter_module_idx:
                continue

            prv_f.append(hidden_states)
            # pop res hidden states
            res_hidden_states = res_hidden_states_tuple[-1]
            res_hidden_states_tuple = res_hidden_states_tuple[:-1]
            hidden_states = torch.cat([hidden_states, res_hidden_states], dim=1)

            if self.training and self.gradient_checkpointing:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs)

                    return custom_forward

                if is_torch_version(">=", "1.11.0"):
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet),
                        hidden_states,
                        temb,
                        image_only_indicator,
                        use_reentrant=False,
                    )
                else:
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet),
                        hidden_states,
                        temb,
                        image_only_indicator,
                    )
            else:
                hidden_states = resnet(
                    hidden_states, temb, image_only_indicator=image_only_indicator,
                )

        if self.upsamplers is not None:
            for upsampler in self.upsamplers:
                hidden_states = upsampler(hidden_states)

        return hidden_states, prv_f


class CrossAttnUpBlockSpatioTemporal(diffusers_unet_3d_blocks.CrossAttnUpBlockSpatioTemporal):
    def forward(
        self,
        hidden_states: torch.FloatTensor,
        res_hidden_states_tuple: Tuple[torch.FloatTensor, ...],
        temb: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        image_only_indicator: Optional[torch.Tensor] = None,
        enter_module_idx: Optional[int] = None,
    ) -> torch.FloatTensor:
        # print("enter_module_idx:", enter_module_idx)
        prv_f = []
        for idx, (resnet, attn) in enumerate(zip(self.resnets, self.attentions)):
            if enter_module_idx is not None and idx < enter_module_idx:
                continue

            prv_f.append(hidden_states)
            # pop res hidden states
            res_hidden_states = res_hidden_states_tuple[-1]
            res_hidden_states_tuple = res_hidden_states_tuple[:-1]
            hidden_states = torch.cat([hidden_states, res_hidden_states], dim=1)

            if self.training and self.gradient_checkpointing:  # TODO

                def create_custom_forward(module, return_dict=None):
                    def custom_forward(*inputs):
                        if return_dict is not None:
                            return module(*inputs, return_dict=return_dict)
                        else:
                            return module(*inputs)

                    return custom_forward

                ckpt_kwargs: Dict[str, Any] = {
                    "use_reentrant": False
                } if is_torch_version(">=", "1.11.0") else {}
                hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(resnet),
                    hidden_states,
                    temb,
                    image_only_indicator,
                    **ckpt_kwargs,
                )
                hidden_states = attn(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    image_only_indicator=image_only_indicator,
                    return_dict=False,
                )[0]
            else:
                hidden_states = resnet(
                    hidden_states, temb, image_only_indicator=image_only_indicator,
                )
                hidden_states = attn(
                    hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    image_only_indicator=image_only_indicator,
                    return_dict=False,
                )[0]

        if self.upsamplers is not None:
            for upsampler in self.upsamplers:
                hidden_states = upsampler(hidden_states)

        return hidden_states, prv_f


update_cls = {
    "DownBlockSpatioTemporal": DownBlockSpatioTemporal,
    "CrossAttnDownBlockSpatioTemporal": CrossAttnDownBlockSpatioTemporal,
    "UpBlockSpatioTemporal": UpBlockSpatioTemporal,
    "CrossAttnUpBlockSpatioTemporal": CrossAttnUpBlockSpatioTemporal,
}

if diffusers_version >= diffusers_0260_v:
    src_get_down_block = diffusers.models.unets.unet_3d_blocks.get_down_block
    src_get_up_block = diffusers.models.unets.unet_3d_blocks.get_up_block
else:
    src_get_down_block = diffusers.models.unet_3d_blocks.get_down_block
    src_get_up_block = diffusers.models.unet_3d_blocks.get_up_block

down_globals = {k: v for k, v in src_get_down_block.__globals__.items()}
down_globals.update(update_cls)
get_down_block = types.FunctionType(
    src_get_down_block.__code__, down_globals, argdefs=src_get_down_block.__defaults__
)


up_globals = {k: v for k, v in src_get_up_block.__globals__.items()}
up_globals.update(update_cls)
get_up_block = types.FunctionType(
    src_get_up_block.__code__, up_globals, argdefs=src_get_up_block.__defaults__
)



from .lora import (
    load_and_fuse_lora,
    unfuse_lora,
    set_and_fuse_adapters,
    delete_adapters,
    get_active_adapters,
)

from onediff.infer_compiler.backends.oneflow.param_utils import update_graph_with_constant_folding_info



from packaging import version
from typing import Union, Dict
from collections import defaultdict

import torch
from onediff.infer_compiler import DeployableModule
from onediff.utils import logger
from diffusers.models.lora import (
    LoRACompatibleConv,
    LoRACompatibleLinear,
)
from .utils import fuse_lora, get_adapter_names, is_peft_available
from diffusers.utils import is_accelerate_available

if is_peft_available():
    import peft

if is_accelerate_available():
    from accelerate.hooks import AlignDevicesHook, CpuOffload, remove_hook_from_module
USE_PEFT_BACKEND = False


def load_lora_into_unet(
    self,
    state_dict,
    network_alphas,
    unet,
    low_cpu_mem_usage=None,
    adapter_name=None,
    _pipeline=None,
    *,
    lora_scale: float = 1.0,
    offload_device="cpu",
    use_cache=False,
):
    if adapter_name is None:
        adapter_name = get_adapter_names(unet)

    if hasattr(unet, "adapter_names"):
        if adapter_name in unet.adapter_names:
            raise ValueError(
                f"[OneDiffX load_lora_into_unet] The adapter name {adapter_name} already exists in UNet"
            )
        else:
            unet.adapter_name.add(adapter_name)
            unet.active_adapter_name[adapter_name] = 1.0
    else:
        unet.adapter_name = set([adapter_name])
        unet.active_adapter_name = {adapter_name: 1.0}

    keys = list(state_dict.keys())
    cls = type(self)

    if all(
        key.startswith(cls.unet_name) or key.startswith(cls.text_encoder_name)
        for key in keys
    ):
        # Load the layers corresponding to UNet.
        logger.info(f"Loading {cls.unet_name}.")

        unet_keys = [k for k in keys if k.startswith(cls.unet_name)]
        state_dict = {
            k.replace(f"{cls.unet_name}.", ""): v
            for k, v in state_dict.items()
            if k in unet_keys
        }

        if network_alphas is not None:
            alpha_keys = [
                k for k in network_alphas.keys() if k.startswith(cls.unet_name)
            ]
            network_alphas = {
                k.replace(f"{cls.unet_name}.", ""): v
                for k, v in network_alphas.items()
                if k in alpha_keys
            }

    else:
        # Otherwise, we're dealing with the old format. This means the `state_dict` should only
        # contain the module names of the `unet` as its keys WITHOUT any prefix.
        warn_message = "You have saved the LoRA weights using the old format. To convert the old LoRA weights to the new format, you can first load them in a dictionary and then create a new dictionary like the following: `new_state_dict = {f'unet.{module_name}': params for module_name, params in old_state_dict.items()}`."
        logger.warning(warn_message)

    # unet.load_attn
    _load_attn_procs(
        unet,
        state_dict,
        network_alphas=network_alphas,
        low_cpu_mem_usage=low_cpu_mem_usage,
        adapter_name=adapter_name,
        _pipeline=_pipeline,
        lora_scale=lora_scale,
        offload_device=offload_device,
        use_cache=use_cache,
    )


# The code is referenced from https://github.com/huggingface/diffusers/blob/ce9825b56bd8a6849e68b9590022e935400659e6/src/diffusers/loaders/unet.py#L383
def _convert_state_dict_legacy_attn_format(self, state_dict, network_alphas):
    is_new_lora_format = all(
        key.startswith(self.unet_name) or key.startswith(self.text_encoder_name)
        for key in state_dict.keys()
    )
    if is_new_lora_format:
        # Strip the `"unet"` prefix.
        is_text_encoder_present = any(
            key.startswith(self.text_encoder_name) for key in state_dict.keys()
        )
        if is_text_encoder_present:
            warn_message = "The state_dict contains LoRA params corresponding to the text encoder which are not being used here. To use both UNet and text encoder related LoRA params, use [`pipe.load_lora_weights()`](https://huggingface.co/docs/diffusers/main/en/api/loaders#diffusers.loaders.LoraLoaderMixin.load_lora_weights)."
            logger.warning(warn_message)
        unet_keys = [k for k in state_dict.keys() if k.startswith(self.unet_name)]
        state_dict = {
            k.replace(f"{self.unet_name}.", ""): v
            for k, v in state_dict.items()
            if k in unet_keys
        }

    # change processor format to 'pure' LoRACompatibleLinear format
    if any("processor" in k.split(".") for k in state_dict.keys()):

        def format_to_lora_compatible(key):
            if "processor" not in key.split("."):
                return key
            return (
                key.replace(".processor", "")
                .replace("to_out_lora", "to_out.0.lora")
                .replace("_lora", ".lora")
            )

        state_dict = {format_to_lora_compatible(k): v for k, v in state_dict.items()}

        if network_alphas is not None:
            network_alphas = {
                format_to_lora_compatible(k): v for k, v in network_alphas.items()
            }
    return state_dict, network_alphas


def _load_attn_procs(
    self,
    pretrained_model_name_or_path_or_dict: Union[str, Dict[str, torch.Tensor]],
    **kwargs,
):

    lora_scale = kwargs.pop("lora_scale", 1.0)
    offload_device = kwargs.pop("offload_device", "cpu")
    use_cache = kwargs.pop("use_cache", False)
    _pipeline = kwargs.pop("_pipeline", None)
    network_alphas = kwargs.pop("network_alphas", None)
    adapter_name = kwargs.pop("adapter_name", None)
    state_dict = pretrained_model_name_or_path_or_dict

    is_network_alphas_none = network_alphas is None

    is_lora = (
        all(("lora" in k or k.endswith(".alpha")) for k in state_dict.keys())
        and not USE_PEFT_BACKEND
    )

    if is_lora:
        # correct keys
        if hasattr(self, "convert_state_dict_legacy_attn_format"):
            state_dict, network_alphas = self.convert_state_dict_legacy_attn_format(
                state_dict, network_alphas
            )
        else:
            state_dict, network_alphas = _convert_state_dict_legacy_attn_format(
                self, state_dict, network_alphas
            )

        if network_alphas is not None:
            network_alphas_keys = list(network_alphas.keys())
            used_network_alphas_keys = set()

        lora_grouped_dict = defaultdict(dict)
        mapped_network_alphas = {}

        all_keys = list(state_dict.keys())
        for key in all_keys:
            value = state_dict.pop(key)
            attn_processor_key, sub_key = (
                ".".join(key.split(".")[:-3]),
                ".".join(key.split(".")[-3:]),
            )
            lora_grouped_dict[attn_processor_key][sub_key] = value

            # Create another `mapped_network_alphas` dictionary so that we can properly map them.
            if network_alphas is not None:
                for k in network_alphas_keys:
                    if k.replace(".alpha", "") in key:
                        mapped_network_alphas.update(
                            {attn_processor_key: network_alphas.get(k)}
                        )
                        used_network_alphas_keys.add(k)

        if not is_network_alphas_none:
            if len(set(network_alphas_keys) - used_network_alphas_keys) > 0:
                raise ValueError(
                    f"[OneDiffX _load_attn_procs] The `network_alphas` has to be empty at this point but has the following keys \n\n {', '.join(network_alphas.keys())}"
                )

        if len(state_dict) > 0:
            raise ValueError(
                f"[OneDiffX _load_attn_procs] The `state_dict` has to be empty at this point but has the following keys \n\n {', '.join(state_dict.keys())}"
            )

        for key, value_dict in lora_grouped_dict.items():
            if isinstance(self, DeployableModule):
                attn_processor = self._torch_module
            else:
                attn_processor = self
            for sub_key in key.split("."):
                attn_processor = getattr(attn_processor, sub_key)

            # Process non-attention layers, which don't have to_{k,v,q,out_proj}_lora layers
            # or add_{k,v,q,out_proj}_proj_lora layers.
            rank = value_dict["lora.down.weight"].shape[0]

            if isinstance(
                attn_processor,
                (
                    LoRACompatibleConv,
                    torch.nn.Conv2d,
                    LoRACompatibleLinear,
                    torch.nn.Linear,
                ),
            ):
                fuse_lora(
                    attn_processor,
                    value_dict,
                    lora_scale,
                    mapped_network_alphas.get(key),
                    rank,
                    offload_device=offload_device,
                    adapter_name=adapter_name,
                )
            elif is_peft_available() and isinstance(
                attn_processor,
                (peft.tuners.lora.layer.Linear, peft.tuners.lora.layer.Conv2d),
            ):
                fuse_lora(
                    attn_processor.base_layer,
                    value_dict,
                    lora_scale,
                    mapped_network_alphas.get(key),
                    rank,
                    offload_device=offload_device,
                    adapter_name=adapter_name,
                )
            else:
                raise ValueError(
                    f"[OneDiffX _load_attn_procs] Module {key} is not a Conv2d or Linear module, got type {type(attn_processor)}"
                )
    else:
        raise ValueError(
            f"[OneDiffX _load_attn_procs] {pretrained_model_name_or_path_or_dict} does not seem to be in the correct format expected by LoRA training."
        )

    is_model_cpu_offload = False
    is_sequential_cpu_offload = False

    if not USE_PEFT_BACKEND:
        if _pipeline is not None:
            for _, component in _pipeline.components.items():
                if isinstance(component, torch.nn.Module) and hasattr(
                    component, "_hf_hook"
                ):
                    is_model_cpu_offload = isinstance(
                        getattr(component, "_hf_hook"), CpuOffload
                    )
                    is_sequential_cpu_offload = isinstance(
                        getattr(component, "_hf_hook"), AlignDevicesHook
                    )

                    logger.info(
                        "Accelerate hooks detected. Since you have called `load_lora_weights()`, the previous hooks will be first removed. Then the LoRA parameters will be loaded and the hooks will be applied again."
                    )
                    remove_hook_from_module(
                        component, recurse=is_sequential_cpu_offload
                    )

        # self.to(dtype=self.dtype, device=self.device)

        # Offload back.
        if is_model_cpu_offload:
            _pipeline.enable_model_cpu_offload()
        elif is_sequential_cpu_offload:
            _pipeline.enable_sequential_cpu_offload()
        # Unsafe code />



from dataclasses import dataclass
from typing import Optional, Tuple, Union

from packaging import version
import importlib.metadata
from oneflow.nn.graph.proxy import ProxyModule

diffusers_0260_v = version.parse("0.26.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

if diffusers_version >= diffusers_0260_v:
    from diffusers.models.unets.unet_spatio_temporal_condition import (
        UNetSpatioTemporalConditionModel as DiffusersUNetSpatioTemporalConditionModel,
    )
else:
    from diffusers.models.unet_spatio_temporal_condition import (
        UNetSpatioTemporalConditionModel as DiffusersUNetSpatioTemporalConditionModel,
    )

import torch

from diffusers.utils import BaseOutput, logging
from diffusers.models.modeling_utils import ModelMixin

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name


@dataclass
class UNetSpatioTemporalConditionOutput(BaseOutput):
    """
    The output of [`UNetSpatioTemporalConditionModel`].

    Args:
        sample (`torch.FloatTensor` of shape `(batch_size, num_frames, num_channels, height, width)`):
            The hidden states output conditioned on `encoder_hidden_states` input. Output of last layer of model.
    """

    sample: torch.FloatTensor = None


class UNetSpatioTemporalConditionModel(DiffusersUNetSpatioTemporalConditionModel):
    def forward(
        self,
        sample: torch.FloatTensor,
        timestep: Union[torch.Tensor, float, int],
        encoder_hidden_states: torch.Tensor,
        added_time_ids: torch.Tensor,
        return_dict: bool = True,
        cache_features: Optional[torch.Tensor] = None,
        cache_branch: Optional[int] = None,
    ) -> Union[UNetSpatioTemporalConditionOutput, Tuple]:
        r"""
        The [`UNetSpatioTemporalConditionModel`] forward method.

        Args:
            sample (`torch.FloatTensor`):
                The noisy input tensor with the following shape `(batch, num_frames, channel, height, width)`.
            timestep (`torch.FloatTensor` or `float` or `int`): The number of timesteps to denoise an input.
            encoder_hidden_states (`torch.FloatTensor`):
                The encoder hidden states with shape `(batch, sequence_length, cross_attention_dim)`.
            added_time_ids: (`torch.FloatTensor`):
                The additional time ids with shape `(batch, num_additional_ids)`. These are encoded with sinusoidal
                embeddings and added to the time embeddings.
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a [`~models.unet_slatio_temporal.UNetSpatioTemporalConditionOutput`] instead of a plain
                tuple.
        Returns:
            [`~models.unet_slatio_temporal.UNetSpatioTemporalConditionOutput`] or `tuple`:
                If `return_dict` is True, an [`~models.unet_slatio_temporal.UNetSpatioTemporalConditionOutput`] is returned, otherwise
                a `tuple` is returned where the first element is the sample tensor.
        """
        # 1. time
        timesteps = timestep
        if not torch.is_tensor(timesteps):
            # TODO: this requires sync between CPU and GPU. So try to pass timesteps as tensors if you can
            # This would be a good case for the `match` statement (Python 3.10+)
            is_mps = sample.device.type == "mps"
            if isinstance(timestep, float):
                dtype = torch.float32 if is_mps else torch.float64
            else:
                dtype = torch.int32 if is_mps else torch.int64
            timesteps = torch.tensor([timesteps], dtype=dtype, device=sample.device)
        elif len(timesteps.shape) == 0:
            timesteps = timesteps[None].to(sample.device)

        # broadcast to batch dimension in a way that's compatible with ONNX/Core ML
        batch_size, num_frames = sample.shape[:2]
        timesteps = timesteps.expand(batch_size)

        t_emb = self.time_proj(timesteps)

        # `Timesteps` does not contain any weights and will always return f32 tensors
        # but time_embedding might actually be running in fp16. so we need to cast here.
        # there might be better ways to encapsulate this.
        t_emb = t_emb.to(dtype=sample.dtype)

        emb = self.time_embedding(t_emb)

        time_embeds = self.add_time_proj(added_time_ids.flatten())
        time_embeds = time_embeds.reshape((batch_size, -1))
        time_embeds = time_embeds.to(emb.dtype)
        aug_emb = self.add_embedding(time_embeds)
        emb = emb + aug_emb

        # Flatten the batch and frames dimensions
        # sample: [batch, frames, channels, height, width] -> [batch * frames, channels, height, width]
        sample = sample.flatten(0, 1)
        # Repeat the embeddings num_video_frames times
        # emb: [batch, channels] -> [batch * frames, channels]
        emb = emb.repeat_interleave(num_frames, dim=0)
        # encoder_hidden_states: [batch, 1, channels] -> [batch * frames, 1, channels]
        encoder_hidden_states = encoder_hidden_states.repeat_interleave(
            num_frames, dim=0
        )

        # 2. pre-process
        sample = self.conv_in(sample)

        image_only_indicator = torch.zeros(
            batch_size, num_frames, dtype=sample.dtype, device=sample.device
        )

        # Branch: 4 down_blocks, each with 3 skip connections. Here we ignore the first skip branch, whose computations only has up_blocks but without down_blocks.
        if cache_branch is not None:
            each_module_num = len(self.down_blocks[0].resnets) + 1
            down_cache_block_idx = cache_branch // each_module_num
            down_cache_module_idx = cache_branch % each_module_num

            up_cache_block_idx = len(self.up_blocks) - 1 - down_cache_block_idx
            up_cache_module_idx = 1 - down_cache_module_idx
            if down_cache_module_idx == each_module_num - 1:
                up_cache_block_idx -= 1
                up_cache_module_idx = 2

        # 3. down
        down_block_res_samples = (sample,)
        for downsample_block in self.down_blocks:
            if (
                hasattr(downsample_block, "has_cross_attention")
                and downsample_block.has_cross_attention
            ):
                sample, res_samples = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    encoder_hidden_states=encoder_hidden_states,
                    image_only_indicator=image_only_indicator,
                )
            else:
                sample, res_samples = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    image_only_indicator=image_only_indicator,
                )

            down_block_res_samples += res_samples

        # 4. mid
        sample = self.mid_block(
            hidden_states=sample,
            temb=emb,
            encoder_hidden_states=encoder_hidden_states,
            image_only_indicator=image_only_indicator,
        )

        # 5. up
        for i, upsample_block in enumerate(self.up_blocks):
            res_samples = down_block_res_samples[-len(upsample_block.resnets) :]
            down_block_res_samples = down_block_res_samples[
                : -len(upsample_block.resnets)
            ]

            if (
                hasattr(upsample_block, "has_cross_attention")
                and upsample_block.has_cross_attention
            ):
                sample, current_record_f = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    encoder_hidden_states=encoder_hidden_states,
                    image_only_indicator=image_only_indicator,
                )
            else:
                sample, current_record_f = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    image_only_indicator=image_only_indicator,
                )

            if cache_branch is not None and i == up_cache_block_idx:
                cache_features = current_record_f[up_cache_module_idx]

        # 6. post-process
        sample = self.conv_norm_out(sample)
        sample = self.conv_act(sample)
        sample = self.conv_out(sample)

        # 7. Reshape back to original shape
        if isinstance(self, ProxyModule):
            # Rewrite for onediff SVD dynamic shape
            sample = sample.unflatten(0, shape=(batch_size, -1))
        else:
            sample = sample.reshape(batch_size, num_frames, *sample.shape[1:])

        if not return_dict:
            return (sample, cache_features)

        return UNetSpatioTemporalConditionOutput(sample=sample)



# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Optional, Tuple

import types
import torch
from oneflow.nn.graph.proxy import ProxyModule

from packaging import version
import importlib.metadata

diffusers_0210_v = version.parse("0.21.0")
diffusers_0260_v = version.parse("0.26.0")
diffusers_0270_v = version.parse("0.27.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

import diffusers
from diffusers.utils import is_torch_version, logging

if diffusers_version >= diffusers_0260_v:
    from diffusers.models.unets import unet_2d_blocks as diffusers_unet_2d_blocks
else:
    from diffusers.models import unet_2d_blocks as diffusers_unet_2d_blocks

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name


if diffusers_version >= diffusers_0210_v:
    class CrossAttnDownBlock2D(diffusers_unet_2d_blocks.CrossAttnDownBlock2D):
        def forward(
            self,
            hidden_states: torch.FloatTensor,
            temb: Optional[torch.FloatTensor] = None,
            encoder_hidden_states: Optional[torch.FloatTensor] = None,
            attention_mask: Optional[torch.FloatTensor] = None,
            cross_attention_kwargs: Optional[Dict[str, Any]] = None,
            encoder_attention_mask: Optional[torch.FloatTensor] = None,
            exist_block_number=None,
            additional_residuals=None,
        ):
            # print("exist_block_number:", exist_block_number, type(self))
            output_states = ()

            lora_scale = (
                cross_attention_kwargs.get("scale", 1.0)
                if cross_attention_kwargs is not None
                else 1.0
            )

            blocks = list(zip(self.resnets, self.attentions))

            for i, (resnet, attn) in enumerate(blocks):
                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module, return_dict=None):
                        def custom_forward(*inputs):
                            if return_dict is not None:
                                return module(*inputs, return_dict=return_dict)
                            else:
                                return module(*inputs)

                        return custom_forward

                    ckpt_kwargs: Dict[str, Any] = {
                        "use_reentrant": False
                    } if is_torch_version(">=", "1.11.0") else {}
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet), hidden_states, temb, **ckpt_kwargs,
                    )
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]
                else:
                    hidden_states = resnet(hidden_states, temb, scale=lora_scale)
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]

                # apply additional residuals to the output of the last pair of resnet and attention blocks
                if i == len(blocks) - 1 and additional_residuals is not None:
                    hidden_states = hidden_states + additional_residuals

                output_states = output_states + (hidden_states,)
                if (
                    exist_block_number is not None
                    and len(output_states) == exist_block_number + 1
                ):
                    return hidden_states, output_states

            if self.downsamplers is not None:
                for downsampler in self.downsamplers:
                    hidden_states = downsampler(hidden_states, scale=lora_scale)

                output_states = output_states + (hidden_states,)

            return hidden_states, output_states


    class DownBlock2D(diffusers_unet_2d_blocks.DownBlock2D):
        def forward(
            self, hidden_states, temb=None, scale: float = 1.0, exist_block_number=None,
        ):
            # print("exist_block_number:", exist_block_number, type(self))
            output_states = ()

            for resnet in self.resnets:
                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module):
                        def custom_forward(*inputs):
                            return module(*inputs)

                        return custom_forward

                    if is_torch_version(">=", "1.11.0"):
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet),
                            hidden_states,
                            temb,
                            use_reentrant=False,
                        )
                    else:
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet), hidden_states, temb
                        )
                else:
                    hidden_states = resnet(hidden_states, temb, scale=scale)

                output_states = output_states + (hidden_states,)
                if (
                    exist_block_number is not None
                    and len(output_states) == exist_block_number + 1
                ):
                    return hidden_states, output_states

            if self.downsamplers is not None:
                for downsampler in self.downsamplers:
                    hidden_states = downsampler(hidden_states, scale=scale)

                output_states = output_states + (hidden_states,)

            return hidden_states, output_states


    class CrossAttnUpBlock2D(diffusers_unet_2d_blocks.CrossAttnUpBlock2D):
        def forward(
            self,
            hidden_states: torch.FloatTensor,
            res_hidden_states_tuple: Tuple[torch.FloatTensor, ...],
            temb: Optional[torch.FloatTensor] = None,
            encoder_hidden_states: Optional[torch.FloatTensor] = None,
            cross_attention_kwargs: Optional[Dict[str, Any]] = None,
            upsample_size: Optional[int] = None,
            output_like: Optional[torch.FloatTensor] = None,
            attention_mask: Optional[torch.FloatTensor] = None,
            encoder_attention_mask: Optional[torch.FloatTensor] = None,
            enter_block_number: Optional[int] = None,
        ):
            # print("enter_block_number:", enter_block_number, type(self))
            prv_f = []
            lora_scale = (
                cross_attention_kwargs.get("scale", 1.0)
                if cross_attention_kwargs is not None
                else 1.0
            )

            for i, (resnet, attn) in enumerate(zip(self.resnets, self.attentions)):
                # pop res hidden states

                if (
                    enter_block_number is not None
                    and i < len(self.resnets) - enter_block_number - 1
                ):
                    continue

                res_hidden_states = res_hidden_states_tuple[-1]
                res_hidden_states_tuple = res_hidden_states_tuple[:-1]

                prv_f.append(hidden_states)
                hidden_states = torch.cat([hidden_states, res_hidden_states], dim=1)

                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module, return_dict=None):
                        def custom_forward(*inputs):
                            if return_dict is not None:
                                return module(*inputs, return_dict=return_dict)
                            else:
                                return module(*inputs)

                        return custom_forward

                    ckpt_kwargs: Dict[str, Any] = {
                        "use_reentrant": False
                    } if is_torch_version(">=", "1.11.0") else {}
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet), hidden_states, temb, **ckpt_kwargs,
                    )
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]
                else:
                    hidden_states = resnet(hidden_states, temb, scale=lora_scale)
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]

            if self.upsamplers is not None:
                for upsampler in self.upsamplers:
                    if isinstance(self, ProxyModule):
                        hidden_states = upsampler(
                            hidden_states, upsample_size, output_like=output_like, scale=lora_scale
                        )
                    else:
                        hidden_states = upsampler(
                            hidden_states, upsample_size, scale=lora_scale
                        )

            return hidden_states, prv_f


    class UpBlock2D(diffusers_unet_2d_blocks.UpBlock2D):
        def forward(
            self,
            hidden_states,
            res_hidden_states_tuple,
            temb=None,
            upsample_size=None,
            output_like: Optional[torch.FloatTensor] = None,
            scale: float = 1.0,
            enter_block_number: Optional[int] = None,
        ):
            # print("enter_block_number:", enter_block_number, type(self))
            prv_f = []

            for idx, resnet in enumerate(self.resnets):

                if (
                    enter_block_number is not None
                    and idx < len(self.resnets) - enter_block_number - 1
                ):
                    continue

                # pop res hidden states
                res_hidden_states = res_hidden_states_tuple[-1]
                res_hidden_states_tuple = res_hidden_states_tuple[:-1]
                prv_f.append(hidden_states)
                hidden_states = torch.cat([hidden_states, res_hidden_states], dim=1)

                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module):
                        def custom_forward(*inputs):
                            return module(*inputs)

                        return custom_forward

                    if is_torch_version(">=", "1.11.0"):
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet),
                            hidden_states,
                            temb,
                            use_reentrant=False,
                        )
                    else:
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet), hidden_states, temb
                        )
                else:
                    hidden_states = resnet(hidden_states, temb, scale=scale)

            if self.upsamplers is not None:
                for upsampler in self.upsamplers:
                    if isinstance(self, ProxyModule):
                        hidden_states = upsampler(hidden_states, upsample_size, scale=scale, output_like=output_like,)
                    else:
                        hidden_states = upsampler(hidden_states, upsample_size, scale=scale,)

            return hidden_states, prv_f
else:

    class CrossAttnDownBlock2D(diffusers_unet_2d_blocks.CrossAttnDownBlock2D):
        def forward(
            self,
            hidden_states: torch.FloatTensor,
            temb: Optional[torch.FloatTensor] = None,
            encoder_hidden_states: Optional[torch.FloatTensor] = None,
            attention_mask: Optional[torch.FloatTensor] = None,
            cross_attention_kwargs: Optional[Dict[str, Any]] = None,
            encoder_attention_mask: Optional[torch.FloatTensor] = None,
            exist_block_number=None,
            additional_residuals=None,
        ):
            # print("exist_block_number:", exist_block_number, type(self))
            if diffusers_version >= diffusers_0270_v:
                if cross_attention_kwargs is not None:
                    if cross_attention_kwargs.get("scale", None) is not None:
                        logger.warning("Passing `scale` to `cross_attention_kwargs` is depcrecated. `scale` will be ignored.")

            output_states = ()

            blocks = list(zip(self.resnets, self.attentions))

            for i, (resnet, attn) in enumerate(blocks):
                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module, return_dict=None):
                        def custom_forward(*inputs):
                            if return_dict is not None:
                                return module(*inputs, return_dict=return_dict)
                            else:
                                return module(*inputs)

                        return custom_forward

                    ckpt_kwargs: Dict[str, Any] = {
                        "use_reentrant": False
                    } if is_torch_version(">=", "1.11.0") else {}
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet), hidden_states, temb, **ckpt_kwargs,
                    )
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]
                else:
                    hidden_states = resnet(hidden_states, temb)
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]

                # apply additional residuals to the output of the last pair of resnet and attention blocks
                if i == len(blocks) - 1 and additional_residuals is not None:
                    hidden_states = hidden_states + additional_residuals

                output_states = output_states + (hidden_states,)
                if (
                    exist_block_number is not None
                    and len(output_states) == exist_block_number + 1
                ):
                    return hidden_states, output_states

            if self.downsamplers is not None:
                for downsampler in self.downsamplers:
                    hidden_states = downsampler(hidden_states)

                output_states = output_states + (hidden_states,)

            return hidden_states, output_states


    class DownBlock2D(diffusers_unet_2d_blocks.DownBlock2D):
        def forward(
            self, hidden_states, temb=None, exist_block_number=None,
        ):
            # print("exist_block_number:", exist_block_number, type(self))
            output_states = ()

            for resnet in self.resnets:
                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module):
                        def custom_forward(*inputs):
                            return module(*inputs)

                        return custom_forward

                    if is_torch_version(">=", "1.11.0"):
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet),
                            hidden_states,
                            temb,
                            use_reentrant=False,
                        )
                    else:
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet), hidden_states, temb
                        )
                else:
                    hidden_states = resnet(hidden_states, temb)

                output_states = output_states + (hidden_states,)
                if (
                    exist_block_number is not None
                    and len(output_states) == exist_block_number + 1
                ):
                    return hidden_states, output_states

            if self.downsamplers is not None:
                for downsampler in self.downsamplers:
                    hidden_states = downsampler(hidden_states)

                output_states = output_states + (hidden_states,)

            return hidden_states, output_states


    class CrossAttnUpBlock2D(diffusers_unet_2d_blocks.CrossAttnUpBlock2D):
        def forward(
            self,
            hidden_states: torch.FloatTensor,
            res_hidden_states_tuple: Tuple[torch.FloatTensor, ...],
            temb: Optional[torch.FloatTensor] = None,
            encoder_hidden_states: Optional[torch.FloatTensor] = None,
            cross_attention_kwargs: Optional[Dict[str, Any]] = None,
            upsample_size: Optional[int] = None,
            output_like: Optional[torch.FloatTensor] = None,
            attention_mask: Optional[torch.FloatTensor] = None,
            encoder_attention_mask: Optional[torch.FloatTensor] = None,
            enter_block_number: Optional[int] = None,
        ):
            # print("enter_block_number:", enter_block_number, type(self))
            if diffusers_version >= diffusers_0270_v:
                if cross_attention_kwargs is not None:
                    if cross_attention_kwargs.get("scale", None) is not None:
                        logger.warning("Passing `scale` to `cross_attention_kwargs` is depcrecated. `scale` will be ignored.")
            prv_f = []

            for i, (resnet, attn) in enumerate(zip(self.resnets, self.attentions)):
                # pop res hidden states

                if (
                    enter_block_number is not None
                    and i < len(self.resnets) - enter_block_number - 1
                ):
                    continue

                res_hidden_states = res_hidden_states_tuple[-1]
                res_hidden_states_tuple = res_hidden_states_tuple[:-1]

                prv_f.append(hidden_states)
                hidden_states = torch.cat([hidden_states, res_hidden_states], dim=1)

                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module, return_dict=None):
                        def custom_forward(*inputs):
                            if return_dict is not None:
                                return module(*inputs, return_dict=return_dict)
                            else:
                                return module(*inputs)

                        return custom_forward

                    ckpt_kwargs: Dict[str, Any] = {
                        "use_reentrant": False
                    } if is_torch_version(">=", "1.11.0") else {}
                    hidden_states = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(resnet), hidden_states, temb, **ckpt_kwargs,
                    )
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]
                else:
                    hidden_states = resnet(hidden_states, temb)
                    hidden_states = attn(
                        hidden_states,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                        return_dict=False,
                    )[0]

            if self.upsamplers is not None:
                for upsampler in self.upsamplers:
                    if isinstance(self, ProxyModule):
                        hidden_states = upsampler(
                            hidden_states, upsample_size, output_like
                        )
                    else:
                        hidden_states = upsampler(
                            hidden_states, upsample_size
                        )

            return hidden_states, prv_f


    class UpBlock2D(diffusers_unet_2d_blocks.UpBlock2D):
        def forward(
            self,
            hidden_states,
            res_hidden_states_tuple,
            temb=None,
            upsample_size=None,
            output_like=None,
            enter_block_number: Optional[int] = None,
        ):
            # print("enter_block_number:", enter_block_number, type(self))
            prv_f = []

            for idx, resnet in enumerate(self.resnets):

                if (
                    enter_block_number is not None
                    and idx < len(self.resnets) - enter_block_number - 1
                ):
                    continue

                # pop res hidden states
                res_hidden_states = res_hidden_states_tuple[-1]
                res_hidden_states_tuple = res_hidden_states_tuple[:-1]
                prv_f.append(hidden_states)
                hidden_states = torch.cat([hidden_states, res_hidden_states], dim=1)

                if self.training and self.gradient_checkpointing:

                    def create_custom_forward(module):
                        def custom_forward(*inputs):
                            return module(*inputs)

                        return custom_forward

                    if is_torch_version(">=", "1.11.0"):
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet),
                            hidden_states,
                            temb,
                            use_reentrant=False,
                        )
                    else:
                        hidden_states = torch.utils.checkpoint.checkpoint(
                            create_custom_forward(resnet), hidden_states, temb
                        )
                else:
                    hidden_states = resnet(hidden_states, temb)

            if self.upsamplers is not None:
                for upsampler in self.upsamplers:
                    if isinstance(self, ProxyModule):
                        hidden_states = upsampler(hidden_states, upsample_size, output_like)
                    else:
                        hidden_states = upsampler(hidden_states, upsample_size)

            return hidden_states, prv_f


update_cls = {
  "CrossAttnDownBlock2D": CrossAttnDownBlock2D,
  "DownBlock2D": DownBlock2D,
  "CrossAttnUpBlock2D": CrossAttnUpBlock2D,
  "UpBlock2D": UpBlock2D,
}

if diffusers_version >= diffusers_0260_v:
    src_get_down_block = diffusers.models.unets.unet_2d_blocks.get_down_block
    src_get_up_block = diffusers.models.unets.unet_2d_blocks.get_up_block
else:
    src_get_down_block = diffusers.models.unet_2d_blocks.get_down_block
    src_get_up_block = diffusers.models.unet_2d_blocks.get_up_block

down_globals = {k : v for k, v in src_get_down_block.__globals__.items()}
down_globals.update(update_cls)
get_down_block = types.FunctionType(src_get_down_block.__code__, down_globals, argdefs=src_get_down_block.__defaults__)


up_globals = {k : v for k, v in src_get_up_block.__globals__.items()}
up_globals.update(update_cls)
get_up_block = types.FunctionType(src_get_up_block.__code__, up_globals, argdefs=src_get_up_block.__defaults__)



import os
from typing import Dict, Union, List
from packaging import version
from collections import OrderedDict

import torch
import diffusers

if version.parse(diffusers.__version__) >= version.parse("0.22.0"):
    from diffusers.utils.import_utils import is_peft_available

    if is_peft_available():
        import peft
else:
    is_peft_available = lambda: False

from onediff.infer_compiler.backends.oneflow.param_utils import update_graph_related_tensor

if version.parse(diffusers.__version__) <= version.parse("0.20.0"):
    from diffusers.loaders import PatchedLoraProjection
else:
    from diffusers.models.lora import PatchedLoraProjection
from onediff.infer_compiler.backends.oneflow.dual_module import DualModule

if version.parse(diffusers.__version__) <= version.parse("0.20.0"):
    from diffusers.loaders import PatchedLoraProjection
else:
    from diffusers.models.lora import PatchedLoraProjection


_adapter_layer_names = ()


def init_lora_infos(self: torch.nn.Module):
    self.r = {}
    self.lora_alpha = {}
    self.scaling = {}
    self.lora_A = {}
    self.lora_B = {}
    self.adapter_names = set()
    self.active_adapter_names = {}


def delete_lora_infos(self, adapter_names):
    if adapter_names is None:
        adapter_names = list(self.adapter_names.copy())
    curr_adapter_names = self.adapter_names.copy()
    for adapter_name in adapter_names:
        if adapter_name not in curr_adapter_names:
            continue
        self.r.pop(adapter_name)
        self.lora_alpha.pop(adapter_name)
        self.scaling.pop(adapter_name)
        self.lora_A.pop(adapter_name)
        self.lora_B.pop(adapter_name)
        self.adapter_names.remove(adapter_name)
        if adapter_name in self.active_adapter_names:
            self.active_adapter_names.pop(adapter_name)


def get_adapter_names(self):
    if not hasattr(self, "adapter_names"):
        result = "default_0"
    else:
        if not isinstance(self.adapter_names, set):
            adapter_names = set([self.adapter_names])
        else:
            adapter_names = self.adapter_names
        for i in range(0, 10000):
            result = f"default_{i}"
            if result not in adapter_names:
                break
    return result


def get_delta_weight(
    self: Union[torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d],
    w_up: torch.Tensor,
    w_down: torch.Tensor,
    weight: float,
):
    if weight == 0:
        return torch.zeros_like(self.weight, dtype=self.weight.dtype, device=self.weight.device)

    if isinstance(self, (torch.nn.Linear, PatchedLoraProjection)):
        lora_weight = torch.bmm(w_up[None, :], w_down[None, :])[0]
    elif isinstance(self, torch.nn.Conv2d):
        lora_weight = torch.mm(w_up.flatten(start_dim=1), w_down.flatten(start_dim=1))
        lora_weight = lora_weight.reshape((self.weight.shape))
    else:
        raise TypeError(
            f"[OneDiffX get_delta_weight] Expect type Linear or Conv2d, got {type(self)}"
        )
    if weight != 1.0:
        lora_weight *= weight
    return lora_weight


def offload_tensor(tensor, device):
    cur_device = tensor.device
    if cur_device == device:
        return tensor.clone()
    else:
        return tensor.to(device)


def _set_adapter(self, adapter_names, adapter_weights):
    if not isinstance(self, (torch.nn.Linear, torch.nn.Conv2d, PatchedLoraProjection)):
        raise TypeError(
            f"[OneDiffX _set_adapter] Expect type Linear or Conv2d, got {type(self)}"
        )
    if isinstance(self, PatchedLoraProjection):
        self = self.regular_linear_layer
    if not hasattr(self, "adapter_names"):
        return
    if adapter_weights is None:
        adapter_weights = 1.0
    if isinstance(adapter_weights, float):
        adapter_weights = [adapter_weights,] * len(adapter_names)
    _unfuse_lora(self)

    dtype, device = self.weight.data.dtype, self.weight.data.device

    delta_weight = None
    for adapter, weight in zip(adapter_names, adapter_weights):
        if adapter not in self.adapter_names:
            continue

        self.active_adapter_names[adapter] = weight
        self.scaling[adapter] = weight * self.lora_alpha[adapter] / self.r[adapter]
        w_down = self.lora_A[adapter].float().to(device)
        w_up = self.lora_B[adapter].float().to(device)
        if delta_weight is None:
            delta_weight = get_delta_weight(
                self, w_up, w_down, self.scaling[adapter]
            )
        else:
            delta_weight += get_delta_weight(
                self, w_up, w_down, self.scaling[adapter]
            )

    if delta_weight is not None:
        fused_weight = self.weight.data.float() + delta_weight
        self.weight.data.copy_(fused_weight.to(device=device, dtype=dtype))
        update_graph_related_tensor(self)


def _delete_adapter(self, adapter_names):
    if not isinstance(self, (torch.nn.Linear, torch.nn.Conv2d, PatchedLoraProjection)):
        raise TypeError(
            f"[OneDiffX _delete_adapter] Expect type Linear or Conv2d, got {type(self)}"
        )
    if isinstance(self, PatchedLoraProjection):
        self = self.regular_linear_layer
    if not hasattr(self, "adapter_names"):
        return
    _unfuse_lora(self, adapter_names=adapter_names)
    delete_lora_infos(self, adapter_names)


def fuse_lora(
    self: Union[torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d],
    state_dict: Dict[str, torch.Tensor] = None,
    lora_scale: float = 1.0,
    alpha: float = None,
    rank: float = None,
    *,
    adapter_name=None,
    fuse=True,
    prefix="lora",
    offload_device="cpu",
) -> None:
    r"""
    This will fuse the LoRA weights in `state_dict` into Linear or Conv2d module.

    Parameters:
        self (Union[torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d]):
            Model layer to be fused, must be Linear or PatchedLoraProjection or Conv2d.
        state_dict (Dict[str, torch.Tensor]):
            Dictionary containing LoRA weight.
        lora_scale (float, optional):
            Scaling factor for LoRA weights. Default is 1.0.
        alpha (float, optional):
            Alpha parameter of LoRA weights. Default is None.
        rank (float, optional):
            Rank of LoRA weights. Default is None.
        prefix (str, optional):
            Prefix for up and down weight keys in the LoRA weight dictionary. Default is "lora".
        offload_device (str, optional):
            Offload Device for backuping weight, can be "cpu" or "cuda". Default is "cpu".
    """
    if not isinstance(self, (torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d)):
        if is_peft_available() and isinstance(
            self, (peft.tuners.lora.layer.Linear, peft.tuners.lora.layer.Conv2d)
        ):
            self = self.base_layer
        else:
            raise TypeError(
                f"[OneDiffX fuse_lora] Only Linear and Conv2d can fuse lora, but got type {type(self)}"
            )

    if isinstance(self, DualModule):
        self = self._torch_module
    if isinstance(self, PatchedLoraProjection):
        self = self.regular_linear_layer

    if not hasattr(self, "adapter_names"):
        init_lora_infos(self)

    dtype, device = self.weight.data.dtype, self.weight.data.device
    down_key = prefix + ".down.weight"
    up_key = prefix + ".up.weight"

    w_down = state_dict[down_key].float().to(device)
    w_up = state_dict[up_key].float().to(device)

    adapter_name = adapter_name if adapter_name is not None else get_adapter_names(self)

    if alpha is None:
        alpha = rank

    self.scaling[adapter_name] = lora_scale * alpha / rank
    self.r[adapter_name] = rank
    self.lora_alpha[adapter_name] = alpha
    self.lora_A[adapter_name] = offload_tensor(w_down, offload_device)
    self.lora_B[adapter_name] = offload_tensor(w_up, offload_device)
    self.adapter_names.add(adapter_name)
    self.active_adapter_names[adapter_name] = lora_scale

    if fuse:
        lora_weight = get_delta_weight(self, w_up, w_down, self.scaling[adapter_name])
        fused_weight = self.weight.data.float() + lora_weight
        self.weight.data.copy_(fused_weight.to(device=device, dtype=dtype))
        update_graph_related_tensor(self)


def _unfuse_lora(
    self: Union[torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d],
    adapter_names: Union[str, List[str]] = None,
):
    assert isinstance(self, (torch.nn.Linear, PatchedLoraProjection, torch.nn.Conv2d))
    if not hasattr(self, "adapter_names"):
        return
    if isinstance(self, DualModule):
        self = self._torch_module
    if isinstance(self, PatchedLoraProjection):
        self = self.regular_linear_layer

    fused_weight = self.weight.data
    dtype, device = fused_weight.dtype, fused_weight.device

    delta_weight = None
    if adapter_names is None:
        adapter_names = self.active_adapter_names.copy()

    for name in adapter_names:
        if name not in self.active_adapter_names:
            continue
        w_down = self.lora_A[name].to(device=device).float()
        w_up = self.lora_B[name].to(device).float()
        if delta_weight is None:
            delta_weight = get_delta_weight(self, w_up, w_down, self.scaling[name]).to(
                dtype=dtype, device=device
            )
        else:
            delta_weight += get_delta_weight(self, w_up, w_down, self.scaling[name]).to(
                dtype=dtype, device=device
            )
        self.active_adapter_names.pop(name)

    if delta_weight is not None:
        self.weight.data -= delta_weight
        update_graph_related_tensor(self)


# the code is referenced from https://github.com/huggingface/diffusers/blob/ce9825b56bd8a6849e68b9590022e935400659e6/src/diffusers/loaders/lora_conversion_utils.py#L24
@classmethod
def _maybe_map_sgm_blocks_to_diffusers(
    cls, state_dict, unet_config, delimiter="_", block_slice_pos=5
):
    # 1. get all state_dict_keys
    all_keys = list(state_dict.keys())
    sgm_patterns = ["input_blocks", "middle_block", "output_blocks"]

    # 2. check if needs remapping, if not return original dict
    is_in_sgm_format = False
    for key in all_keys:
        if any(p in key for p in sgm_patterns):
            is_in_sgm_format = True
            break

    if not is_in_sgm_format:
        return state_dict
    # 3. Else remap from SGM patterns
    new_state_dict = {}
    inner_block_map = ["resnets", "attentions", "upsamplers"]

    # Retrieves # of down, mid and up blocks
    input_block_ids, middle_block_ids, output_block_ids = set(), set(), set()

    for layer in all_keys:
        if "text" in layer:
            new_state_dict[layer] = state_dict.pop(layer)
        else:
            layer_id = int(layer.split(delimiter)[:block_slice_pos][-1])
            if sgm_patterns[0] in layer:
                input_block_ids.add(layer_id)
            elif sgm_patterns[1] in layer:
                middle_block_ids.add(layer_id)
            elif sgm_patterns[2] in layer:
                output_block_ids.add(layer_id)
            else:
                raise ValueError(
                    f"Checkpoint not supported because layer {layer} not supported."
                )

    input_blocks = {
        layer_id: [
            key for key in state_dict if f"input_blocks{delimiter}{layer_id}" in key
        ]
        for layer_id in input_block_ids
    }
    middle_blocks = {
        layer_id: [
            key for key in state_dict if f"middle_block{delimiter}{layer_id}" in key
        ]
        for layer_id in middle_block_ids
    }
    output_blocks = {
        layer_id: [
            key for key in state_dict if f"output_blocks{delimiter}{layer_id}" in key
        ]
        for layer_id in output_block_ids
    }

    # Rename keys accordingly
    for i in input_block_ids:
        block_id = (i - 1) // (unet_config.layers_per_block + 1)
        layer_in_block_id = (i - 1) % (unet_config.layers_per_block + 1)

        for key in input_blocks[i]:
            inner_block_id = int(key.split(delimiter)[block_slice_pos])
            inner_block_key = (
                inner_block_map[inner_block_id] if "op" not in key else "downsamplers"
            )
            inner_layers_in_block = str(layer_in_block_id) if "op" not in key else "0"
            new_key = delimiter.join(
                key.split(delimiter)[: block_slice_pos - 1]
                + [str(block_id), inner_block_key, inner_layers_in_block]
                + key.split(delimiter)[block_slice_pos + 1 :]
            )
            new_state_dict[new_key] = state_dict.pop(key)

    for i in middle_block_ids:
        key_part = None
        if i == 0:
            key_part = [inner_block_map[0], "0"]
        elif i == 1:
            key_part = [inner_block_map[1], "0"]
        elif i == 2:
            key_part = [inner_block_map[0], "1"]
        else:
            raise ValueError(f"Invalid middle block id {i}.")

        for key in middle_blocks[i]:
            new_key = delimiter.join(
                key.split(delimiter)[: block_slice_pos - 1]
                + key_part
                + key.split(delimiter)[block_slice_pos:]
            )
            new_state_dict[new_key] = state_dict.pop(key)

    for i in output_block_ids:
        block_id = i // (unet_config.layers_per_block + 1)
        layer_in_block_id = i % (unet_config.layers_per_block + 1)

        for key in output_blocks[i]:
            inner_block_id = int(key.split(delimiter)[block_slice_pos])
            inner_block_key = inner_block_map[inner_block_id]
            inner_layers_in_block = (
                str(layer_in_block_id) if inner_block_id < 2 else "0"
            )
            new_key = delimiter.join(
                key.split(delimiter)[: block_slice_pos - 1]
                + [str(block_id), inner_block_key, inner_layers_in_block]
                + key.split(delimiter)[block_slice_pos + 1 :]
            )
            new_state_dict[new_key] = state_dict.pop(key)

    if len(state_dict) > 0:
        raise ValueError("At this point all state dict entries have to be converted.")

    return new_state_dict



# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple, Union

from packaging import version
import importlib.metadata
from oneflow.nn.graph.proxy import ProxyModule

diffusers_0210_v = version.parse("0.21.0")
diffusers_0260_v = version.parse("0.26.0")
diffusers_0270_v = version.parse("0.27.0")
diffusers_version = version.parse(importlib.metadata.version("diffusers"))

import torch
import diffusers

from diffusers.utils import BaseOutput, logging
from diffusers.models.modeling_utils import ModelMixin

if diffusers_version >= diffusers_0260_v:
    from diffusers.models.unets.unet_2d_condition import UNet2DConditionModel as DiffusersUNet2DConditionModel 
else:
    from diffusers.models.unet_2d_condition import UNet2DConditionModel as DiffusersUNet2DConditionModel

try:
    USE_PEFT_BACKEND = diffusers.utils.USE_PEFT_BACKEND
    scale_lora_layers = diffusers.utils.scale_lora_layers
    unscale_lora_layers = diffusers.utils.unscale_lora_layers
except Exception as e:
    USE_PEFT_BACKEND = False

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name


@dataclass
class UNet2DConditionOutput(BaseOutput):
    """
    The output of [`UNet2DConditionModel`].

    Args:
        sample (`torch.FloatTensor` of shape `(batch_size, num_channels, height, width)`):
            The hidden states output conditioned on `encoder_hidden_states` input. Output of last layer of model.
    """

    sample: torch.FloatTensor = None


class UNet2DConditionModel(DiffusersUNet2DConditionModel):
    def forward(
        self,
        sample: torch.FloatTensor,
        timestep: Union[torch.Tensor, float, int],
        encoder_hidden_states: torch.Tensor,
        class_labels: Optional[torch.Tensor] = None,
        timestep_cond: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        added_cond_kwargs: Optional[Dict[str, torch.Tensor]] = None,
        down_block_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
        mid_block_additional_residual: Optional[torch.Tensor] = None,
        down_intrablock_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        return_dict: bool = True,
        replicate_prv_feature: Optional[List[torch.Tensor]] = None,
        cache_layer_id: Optional[int] = None,
        cache_block_id: Optional[int] = None,
    ) -> Union[UNet2DConditionOutput, Tuple]:
        r"""
        The [`UNet2DConditionModel`] forward method.

        Args:
            sample (`torch.FloatTensor`):
                The noisy input tensor with the following shape `(batch, channel, height, width)`.
            timestep (`torch.FloatTensor` or `float` or `int`): The number of timesteps to denoise an input.
            encoder_hidden_states (`torch.FloatTensor`):
                The encoder hidden states with shape `(batch, sequence_length, feature_dim)`.
            encoder_attention_mask (`torch.Tensor`):
                A cross-attention mask of shape `(batch, sequence_length)` is applied to `encoder_hidden_states`. If
                `True` the mask is kept, otherwise if `False` it is discarded. Mask will be converted into a bias,
                which adds large negative values to the attention scores corresponding to "discard" tokens.
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a [`~models.unet_2d_condition.UNet2DConditionOutput`] instead of a plain
                tuple.
            cross_attention_kwargs (`dict`, *optional*):
                A kwargs dictionary that if specified is passed along to the [`AttnProcessor`].
            added_cond_kwargs: (`dict`, *optional*):
                A kwargs dictionary containin additional embeddings that if specified are added to the embeddings that
                are passed along to the UNet blocks.

        Returns:
            [`~models.unet_2d_condition.UNet2DConditionOutput`] or `tuple`:
                If `return_dict` is True, an [`~models.unet_2d_condition.UNet2DConditionOutput`] is returned, otherwise
                a `tuple` is returned where the first element is the sample tensor.
        """
        # By default samples have to be AT least a multiple of the overall upsampling factor.
        # The overall upsampling factor is equal to 2 ** (# num of upsampling layers).
        # However, the upsampling interpolation output size can be forced to fit any upsampling size
        # on the fly if necessary.
        default_overall_up_factor = 2 ** self.num_upsamplers

        # upsample size should be forwarded when sample is not a multiple of `default_overall_up_factor`
        # forward_upsample_size = False
        # interpolate through upsample_size
        forward_upsample_size = True
        upsample_size = None

        for dim in sample.shape[-2:]:
            if dim % default_overall_up_factor != 0:
                # Forward upsample size to force interpolation output size.
                forward_upsample_size = True
                break

        # ensure attention_mask is a bias, and give it a singleton query_tokens dimension
        # expects mask of shape:
        #   [batch, key_tokens]
        # adds singleton query_tokens dimension:
        #   [batch,                    1, key_tokens]
        # this helps to broadcast it as a bias over attention scores, which will be in one of the following shapes:
        #   [batch,  heads, query_tokens, key_tokens] (e.g. torch sdp attn)
        #   [batch * heads, query_tokens, key_tokens] (e.g. xformers or classic attn)
        if attention_mask is not None:
            # assume that mask is expressed as:
            #   (1 = keep,      0 = discard)
            # convert mask into a bias that can be added to attention scores:
            #       (keep = +0,     discard = -10000.0)
            attention_mask = (1 - attention_mask.to(sample.dtype)) * -10000.0
            attention_mask = attention_mask.unsqueeze(1)

        # convert encoder_attention_mask to a bias the same way we do for attention_mask
        if encoder_attention_mask is not None:
            encoder_attention_mask = (
                1 - encoder_attention_mask.to(sample.dtype)
            ) * -10000.0
            encoder_attention_mask = encoder_attention_mask.unsqueeze(1)

        # 0. center input if necessary
        if self.config.center_input_sample:
            sample = 2 * sample - 1.0

        # 1. time
        if diffusers_version < diffusers_0270_v:
            timesteps = timestep
            if not torch.is_tensor(timesteps):
                # TODO: this requires sync between CPU and GPU. So try to pass timesteps as tensors if you can
                # This would be a good case for the `match` statement (Python 3.10+)
                is_mps = sample.device.type == "mps"
                if isinstance(timestep, float):
                    dtype = torch.float32 if is_mps else torch.float64
                else:
                    dtype = torch.int32 if is_mps else torch.int64
                timesteps = torch.tensor([timesteps], dtype=dtype, device=sample.device)
            elif len(timesteps.shape) == 0:
                timesteps = timesteps[None].to(sample.device)

            # broadcast to batch dimension in a way that's compatible with ONNX/Core ML
            # modified to support dynamic shape for onediff
            if isinstance(self, ProxyModule):
                timesteps = torch._C.broadcast_dim_like(timesteps, sample, dim=0)
            else:
                timesteps = timesteps.expand(sample.shape[0])

            t_emb = self.time_proj(timesteps)

            # `Timesteps` does not contain any weights and will always return f32 tensors
            # but time_embedding might actually be running in fp16. so we need to cast here.
            # there might be better ways to encapsulate this.
            t_emb = t_emb.to(dtype=sample.dtype)
        else:
            t_emb = self.get_time_embed(sample=sample, timestep=timestep)

        emb = self.time_embedding(t_emb, timestep_cond)
        aug_emb = None

        if diffusers_version < diffusers_0270_v:
            if self.class_embedding is not None:
                if class_labels is None:
                    raise ValueError(
                        "class_labels should be provided when num_class_embeds > 0"
                    )

                if self.config.class_embed_type == "timestep":
                    class_labels = self.time_proj(class_labels)

                    # `Timesteps` does not contain any weights and will always return f32 tensors
                    # there might be better ways to encapsulate this.
                    class_labels = class_labels.to(dtype=sample.dtype)

                class_emb = self.class_embedding(class_labels).to(dtype=sample.dtype)

                if self.config.class_embeddings_concat:
                    emb = torch.cat([emb, class_emb], dim=-1)
                else:
                    emb = emb + class_emb

            if self.config.addition_embed_type == "text":
                aug_emb = self.add_embedding(encoder_hidden_states)
            elif self.config.addition_embed_type == "text_image":
                # Kandinsky 2.1 - style
                if "image_embeds" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `addition_embed_type` set to 'text_image' which requires the keyword argument `image_embeds` to be passed in `added_cond_kwargs`"
                    )

                image_embs = added_cond_kwargs.get("image_embeds")
                text_embs = added_cond_kwargs.get("text_embeds", encoder_hidden_states)
                aug_emb = self.add_embedding(text_embs, image_embs)
            elif self.config.addition_embed_type == "text_time":
                # SDXL - style
                if "text_embeds" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `addition_embed_type` set to 'text_time' which requires the keyword argument `text_embeds` to be passed in `added_cond_kwargs`"
                    )
                text_embeds = added_cond_kwargs.get("text_embeds")
                if "time_ids" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `addition_embed_type` set to 'text_time' which requires the keyword argument `time_ids` to be passed in `added_cond_kwargs`"
                    )
                time_ids = added_cond_kwargs.get("time_ids")
                time_embeds = self.add_time_proj(time_ids.flatten())
                # modified to support dynamic shape for onediff
                if isinstance(self, ProxyModule):
                    time_embeds = time_embeds.unflatten(
                        dim=0, shape=(-1, time_ids.shape[1])
                    ).flatten(1, 2)
                else:
                    time_embeds = time_embeds.reshape((text_embeds.shape[0], -1))

                add_embeds = torch.concat([text_embeds, time_embeds], dim=-1)
                add_embeds = add_embeds.to(emb.dtype)
                aug_emb = self.add_embedding(add_embeds)
            elif self.config.addition_embed_type == "image":
                # Kandinsky 2.2 - style
                if "image_embeds" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `addition_embed_type` set to 'image' which requires the keyword argument `image_embeds` to be passed in `added_cond_kwargs`"
                    )
                image_embs = added_cond_kwargs.get("image_embeds")
                aug_emb = self.add_embedding(image_embs)
            elif self.config.addition_embed_type == "image_hint":
                # Kandinsky 2.2 - style
                if (
                    "image_embeds" not in added_cond_kwargs
                    or "hint" not in added_cond_kwargs
                ):
                    raise ValueError(
                        f"{self.__class__} has the config param `addition_embed_type` set to 'image_hint' which requires the keyword arguments `image_embeds` and `hint` to be passed in `added_cond_kwargs`"
                    )
                image_embs = added_cond_kwargs.get("image_embeds")
                hint = added_cond_kwargs.get("hint")
                aug_emb, hint = self.add_embedding(image_embs, hint)
                sample = torch.cat([sample, hint], dim=1)
        
        else:
            aug_emb = self.get_aug_embed(
                emb=emb, encoder_hidden_states=encoder_hidden_states, added_cond_kwargs=added_cond_kwargs
            )
            if self.config.addition_embed_type == "image_hint":
                aug_emb, hint = aug_emb
                sample = torch.cat([sample, hint], dim=1)
        
        emb = emb + aug_emb if aug_emb is not None else emb

        if self.time_embed_act is not None:
            emb = self.time_embed_act(emb)

        if diffusers_version < diffusers_0270_v:
            if (
                self.encoder_hid_proj is not None
                and self.config.encoder_hid_dim_type == "text_proj"
            ):
                encoder_hidden_states = self.encoder_hid_proj(encoder_hidden_states)
            elif (
                self.encoder_hid_proj is not None
                and self.config.encoder_hid_dim_type == "text_image_proj"
            ):
                # Kadinsky 2.1 - style
                if "image_embeds" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `encoder_hid_dim_type` set to 'text_image_proj' which requires the keyword argument `image_embeds` to be passed in  `added_conditions`"
                    )

                image_embeds = added_cond_kwargs.get("image_embeds")
                encoder_hidden_states = self.encoder_hid_proj(
                    encoder_hidden_states, image_embeds
                )
            elif (
                self.encoder_hid_proj is not None
                and self.config.encoder_hid_dim_type == "image_proj"
            ):
                # Kandinsky 2.2 - style
                if "image_embeds" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `encoder_hid_dim_type` set to 'image_proj' which requires the keyword argument `image_embeds` to be passed in  `added_conditions`"
                    )
                image_embeds = added_cond_kwargs.get("image_embeds")
                encoder_hidden_states = self.encoder_hid_proj(image_embeds)
            elif (
                self.encoder_hid_proj is not None
                and self.config.encoder_hid_dim_type == "ip_image_proj"
            ):
                if "image_embeds" not in added_cond_kwargs:
                    raise ValueError(
                        f"{self.__class__} has the config param `encoder_hid_dim_type` set to 'ip_image_proj' which requires the keyword argument `image_embeds` to be passed in  `added_conditions`"
                    )
                image_embeds = added_cond_kwargs.get("image_embeds")
                image_embeds = self.encoder_hid_proj(image_embeds).to(
                    encoder_hidden_states.dtype
                )
                encoder_hidden_states = torch.cat(
                    [encoder_hidden_states, image_embeds], dim=1
                )
        else:
            encoder_hidden_states = self.process_encoder_hidden_states(
                encoder_hidden_states=encoder_hidden_states, added_cond_kwargs=added_cond_kwargs
            )

        # 2. pre-process
        sample = self.conv_in(sample)

        # 2.5 GLIGEN position net
        if (
            cross_attention_kwargs is not None
            and cross_attention_kwargs.get("gligen", None) is not None
        ):
            cross_attention_kwargs = cross_attention_kwargs.copy()
            gligen_args = cross_attention_kwargs.pop("gligen")
            cross_attention_kwargs["gligen"] = {
                "objs": self.position_net(**gligen_args)
            }

        # 3. down
        lora_scale = (
            cross_attention_kwargs.get("scale", 1.0)
            if cross_attention_kwargs is not None
            else 1.0
        )
        if USE_PEFT_BACKEND:
            # weight the lora layers by setting `lora_scale` for each PEFT layer
            scale_lora_layers(self, lora_scale)

        is_controlnet = (
            mid_block_additional_residual is not None
            and down_block_additional_residuals is not None
        )
        # using new arg down_intrablock_additional_residuals for T2I-Adapters, to distinguish from controlnets
        is_adapter = down_intrablock_additional_residuals is not None
        # maintain backward compatibility for legacy usage, where
        #       T2I-Adapter and ControlNet both use down_block_additional_residuals arg
        #       but can only use one or the other
        if (
            not is_adapter
            and mid_block_additional_residual is None
            and down_block_additional_residuals is not None
        ):
            deprecate(
                "T2I should not use down_block_additional_residuals",
                "1.3.0",
                "Passing intrablock residual connections with `down_block_additional_residuals` is deprecated \
                       and will be removed in diffusers 1.3.0.  `down_block_additional_residuals` should only be used \
                       for ControlNet. Please make sure use `down_intrablock_additional_residuals` instead. ",
                standard_warn=False,
            )
            down_intrablock_additional_residuals = down_block_additional_residuals
            is_adapter = True

        down_block_res_samples = (sample,)

        assert replicate_prv_feature is None
        for downsample_block in self.down_blocks:
            if (
                hasattr(downsample_block, "has_cross_attention")
                and downsample_block.has_cross_attention
            ):
                # For t2i-adapter CrossAttnDownBlock2D
                additional_residuals = {}
                if is_adapter and len(down_intrablock_additional_residuals) > 0:
                    additional_residuals[
                        "additional_residuals"
                    ] = down_intrablock_additional_residuals.pop(0)

                sample, res_samples = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=attention_mask,
                    cross_attention_kwargs=cross_attention_kwargs,
                    encoder_attention_mask=encoder_attention_mask,
                    **additional_residuals,
                )
            else:
                if diffusers_version < diffusers_0210_v or diffusers_version >= diffusers_0270_v:
                    sample, res_samples = downsample_block(
                        hidden_states=sample, temb=emb
                    )
                else:
                    sample, res_samples = downsample_block(
                        hidden_states=sample, temb=emb, scale=lora_scale
                    )
                if is_adapter and len(down_intrablock_additional_residuals) > 0:
                    sample += down_intrablock_additional_residuals.pop(0)

            down_block_res_samples += res_samples

        if is_controlnet:
            new_down_block_res_samples = ()

            for down_block_res_sample, down_block_additional_residual in zip(
                down_block_res_samples, down_block_additional_residuals
            ):
                down_block_res_sample = (
                    down_block_res_sample + down_block_additional_residual
                )
                new_down_block_res_samples = new_down_block_res_samples + (
                    down_block_res_sample,
                )

            down_block_res_samples = new_down_block_res_samples

        # 4. mid
        if self.mid_block is not None:
            if (
                hasattr(self.mid_block, "has_cross_attention")
                and self.mid_block.has_cross_attention
            ):
                sample = self.mid_block(
                    sample,
                    emb,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=attention_mask,
                    cross_attention_kwargs=cross_attention_kwargs,
                    encoder_attention_mask=encoder_attention_mask,
                )
            else:
                sample = self.mid_block(sample, emb)

            # To support T2I-Adapter-XL
            if (
                is_adapter
                and len(down_intrablock_additional_residuals) > 0
                and sample.shape == down_intrablock_additional_residuals[0].shape
            ):
                sample += down_intrablock_additional_residuals.pop(0)

        if is_controlnet:
            sample = sample + mid_block_additional_residual

        # 5. up
        if cache_block_id is not None:
            max_block_depth = (
                len(self.down_blocks[cache_layer_id].attentions)
                if hasattr(self.down_blocks[cache_layer_id], "attentions")
                else len(self.down_blocks[cache_layer_id].resnets)
            )
            if cache_block_id == max_block_depth:
                cache_block_id = 0
                cache_layer_id += 1
            else:
                cache_block_id += 1
        # print("down_block_res_samples:", [res_sample.shape for res_sample in down_block_res_samples])
        # print(cache_block_id, cache_layer_id)
        prv_f = None
        for i, upsample_block in enumerate(self.up_blocks):
            is_final_block = i == len(self.up_blocks) - 1

            res_samples = down_block_res_samples[-len(upsample_block.resnets) :]
            down_block_res_samples = down_block_res_samples[
                : -len(upsample_block.resnets)
            ]
            # print(sample.shape, [res_sample.shape for res_sample in res_samples])
            # if we have not reached the final block and need to forward the
            # upsample size, we do it here
            if not is_final_block and forward_upsample_size:
                upsample_size = down_block_res_samples[-1].shape[2:]

            # To support dynamic switching of special resolutions, pass a like tensor.
            output_like = None
            if not is_final_block:
                output_like = down_block_res_samples[-1]

            if (
                hasattr(upsample_block, "has_cross_attention")
                and upsample_block.has_cross_attention
            ):
                sample, current_record_f = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    encoder_hidden_states=encoder_hidden_states,
                    cross_attention_kwargs=cross_attention_kwargs,
                    upsample_size=upsample_size,
                    output_like=output_like,
                    attention_mask=attention_mask,
                    encoder_attention_mask=encoder_attention_mask,
                )
            else:
                if diffusers_version < diffusers_0210_v or diffusers_version >= diffusers_0270_v:
                    sample, current_record_f = upsample_block(
                        hidden_states=sample,
                        temb=emb,
                        res_hidden_states_tuple=res_samples,
                        upsample_size=upsample_size,
                        output_like=output_like,
                    )
                else:
                    sample, current_record_f = upsample_block(
                        hidden_states=sample,
                        temb=emb,
                        res_hidden_states_tuple=res_samples,
                        upsample_size=upsample_size,
                        output_like=output_like,
                        scale=lora_scale,
                    )

            # print(cache_layer_id, current_record_f is None, i == len(self.up_blocks) - cache_layer_id - 1)
            # print("Append prv_feature with shape:", sample.shape)
            if (
                cache_layer_id is not None
                and current_record_f is not None
                and i == len(self.up_blocks) - cache_layer_id - 1
            ):
                prv_f = current_record_f[-cache_block_id - 1]

        # 6. post-process
        if self.conv_norm_out:
            sample = self.conv_norm_out(sample)
            sample = self.conv_act(sample)
        sample = self.conv_out(sample)
        if not return_dict:
            return (
                sample,
                prv_f,
            )

        return UNet2DConditionOutput(sample=sample)
